# -*- coding: utf-8 -*-
"""
Nipype workflow to detect and remove ardifacts in diffusion MRI.
"""
import os.path as op

import nipype.pipeline.engine as pe
from   nipype.interfaces.fsl import BET, ExtractROI
from   nipype.interfaces.utility import Function, IdentityInterface
from   nipype.workflows.dmri.fsl.utils import eddy_rotate_bvecs, b0_average, b0_indices
from   nipype.workflows.dmri.fsl import hmc_pipeline

from ..interfaces import Eddy

from .utils import (dti_acquisition_parameters,
                    nlmeans_denoise, reslice,
                    rapidart_dti_artifact_detection,)

from .._utils  import format_pair_list
from ..config  import setup_node, get_config_setting
from ..utils   import (get_datasink,
                       get_input_node,
                       remove_ext,
                       extend_trait_list,
                       get_input_file_name,
                       extension_duplicates,
                       )


def dti_artifact_correction(wf_name="dti_artifact_correction"):
    """ Run the diffusion MRI pre-processing workflow against the diff files in `data_dir`.

    It will resample/regrid the diffusion image to have isometric voxels.
    Corrects for head motion correction and Eddy currents.
    Estimates motion outliers and exports motion reports using nipype.algorithms.RapidArt.

    Nipype Inputs
    -------------
    dti_art_input.diff: traits.File
        path to the diffusion MRI image

    dti_art_input.bval: traits.File
        path to the bvals file

    dti_art_input.bvec: traits.File
        path to the bvecs file


    Nipype Outputs
    --------------
    dti_art_output.eddy_corr_file: traits.File
        Eddy currents corrected DTI image.

    dti_art_output.bvec_rotated: traits.File
        Rotated bvecs file

    dti_art_output.brain_mask_1: traits.File
        Brain mask extracted using BET on the first B0 image.

    dti_art_output.brain_mask_2: traits.File
        Brain mask extracted using BET on the average B0 image,
        after motion correction.

    dti_art_output.acpq: traits.File
        Text file with acquisition parameters calculated for Eddy.

    dti_art_output.index: traits.File
        Text file with acquisition indices calculated for Eddy.

    dti_art_output.avg_b0: traits.File
        The average b=0 image extracted from the motion and eddy
        currents correted diffusion MRI.

    dti_art_output.hmc_corr_file: traits.File

    dti_art_output.hmc_corr_bvec: traits.File

    dti_art_output.hmc_corr_xfms: traits.File

    dti_art_output.art_displacement_files: traits.File

    dti_art_output.art_intensity_files: traits.File

    dti_art_output.art_norm_files: traits.File

    dti_art_output.art_outlier_files: traits.File

    dti_art_output.art_plot_files: traits.File

    dti_art_output.art_statistic_files: traits.File

    Returns
    -------
    wf: nipype Workflow
    """
    # specify input and output fields
    in_fields  = ["diff", "bval", "bvec"]
    out_fields = ["eddy_corr_file",
                  "bvec_rotated",
                  "brain_mask_1",
                  "brain_mask_2",
                  "acqp",
                  "index",
                  "avg_b0",
                 ]

    do_rapidart = get_config_setting("dmri.artifact_detect", True)
    if do_rapidart:
        out_fields += ["hmc_corr_file",
                       "hmc_corr_bvec",
                       "hmc_corr_xfms",
                       "art_displacement_files",
                       "art_intensity_files",
                       "art_norm_files",
                       "art_outlier_files",
                       "art_plot_files",
                       "art_statistic_files",
                      ]

    # input interface
    dti_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                           name="dti_art_input")

    # resample
    resample = setup_node(Function(function=reslice,
                                   input_names=['in_file', 'new_zooms', 'order', 'out_file'],
                                   output_names=['out_file']),
                          name='dti_reslice')

    ## extract first b0 for Eddy and HMC brain mask
    list_b0 = pe.Node(Function(function=b0_indices,
                               input_names=['in_bval'],
                               output_names=['out_idx'],),
                               name='b0_indices')

    extract_b0 = pe.Node(ExtractROI(t_size=1),
                         name="extract_first_b0")

    # For Eddy, the mask is only used for selecting voxels for the estimation of the hyperparameters,
    # so isn’t very critical.
    # Note also that it is better with a too conservative (small) mask than a too big.
    bet_dwi0 = setup_node(BET(frac=0.3, mask=True, robust=True),
                          name='bet_dwi_pre')

    pick_first = lambda lst: lst[0]

    # motion artifacts detection, requires linear co-registration for motion estimation.
    if do_rapidart:
        # head motion correction
        hmc = hmc_pipeline()

        art = setup_node(rapidart_dti_artifact_detection(), name="detect_artifacts")

    # Eddy
    eddy = setup_node(Eddy(method='jac'), name="eddy")

    ## acquisition parameters for Eddy
    write_acqp = setup_node(Function(function=dti_acquisition_parameters,
                                     input_names=["in_file"],
                                     output_names=["out_acqp", "out_index"],),
                            name="write_acqp")

    ## rotate b-vecs
    rot_bvec = setup_node(Function(function=eddy_rotate_bvecs,
                                   input_names=["in_bvec", "eddy_params"],
                                   output_names=["out_file"],),
                          name="rot_bvec")

    ## extract all b0s and average them after Eddy correction
    avg_b0_post = pe.Node(Function(function=b0_average,
                                   input_names=['in_dwi', 'in_bval'],
                                   output_names=['out_file'],),
                          name='b0_avg_post')

    bet_dwi1 = setup_node(BET(frac=0.3, mask=True, robust=True),
                          name='bet_dwi_post')

    # nlmeans denoise
    apply_nlmeans = get_config_setting("dmri.apply_nlmeans", True)
    if apply_nlmeans:
        nlmeans = setup_node(Function(function=nlmeans_denoise,
                                      input_names=['in_file', 'mask_file', 'out_file', 'N'],
                                      output_names=['out_file']),
                             name='nlmeans_denoise')

    # output interface
    dti_output = setup_node(IdentityInterface(fields=out_fields),
                            name="dti_art_output")

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # Connect the nodes
    wf.connect([
                # resample to iso-voxel
                (dti_input, resample, [("diff", "in_file"),]),

                # read from input file the acquisition parameters for eddy
                (dti_input, write_acqp, [("diff", "in_file")]),

                # reference mask for hmc and eddy
                (dti_input,  list_b0,    [("bval",     "in_bval")]),
                (resample,   extract_b0, [("out_file", "in_file")]),
                (list_b0,    extract_b0, [(("out_idx", pick_first), "t_min")]),

                (extract_b0, bet_dwi0,   [("roi_file", "in_file")]),

                # Eddy
                (resample,   eddy, [("out_file",  "in_file")]),
                (bet_dwi0,   eddy, [("mask_file", "in_mask")]),
                (dti_input,  eddy, [("bval",      "in_bval"),
                                    ("bvec",      "in_bvec")
                                   ]),
                (write_acqp, eddy, [("out_acqp",  "in_acqp"),
                                    ("out_index", "in_index")
                                   ]),

                # rotate bvecs
                (dti_input, rot_bvec, [("bvec",          "in_bvec")]),
                (eddy,      rot_bvec, [("out_parameter", "eddy_params")]),

                # final avg b0
                (dti_input,   avg_b0_post, [("bval",          "in_bval")]),
                (eddy,        avg_b0_post, [("out_corrected", "in_dwi" )]),
                (avg_b0_post, bet_dwi1,    [("out_file",      "in_file")]),

                # output
                (write_acqp,  dti_output,  [("out_acqp",  "acqp"),
                                            ("out_index", "index")]),
                (bet_dwi0,    dti_output,  [("mask_file", "brain_mask_1")]),
                (bet_dwi1,    dti_output,  [("mask_file", "brain_mask_2")]),
                (rot_bvec,    dti_output,  [("out_file",  "bvec_rotated")]),
                (avg_b0_post, dti_output,  [("out_file",  "avg_b0")]),
              ])

    if apply_nlmeans:
        wf.connect([
                    # non-local means
                    (eddy,     nlmeans,   [("out_corrected", "in_file")]),
                    (bet_dwi1, nlmeans,   [("mask_file",     "mask_file")]),

                    # output
                    (nlmeans, dti_output, [("out_file", "eddy_corr_file")]),
                   ])
    else:
        wf.connect([
                    # output
                    (eddy, dti_output, [("out_corrected", "eddy_corr_file")]),
                   ])

    if do_rapidart:
        wf.connect([
                    # head motion correction
                    (dti_input, hmc, [("bval", "inputnode.in_bval"),
                                      ("bvec", "inputnode.in_bvec"),
                                     ]),
                    (resample,  hmc, [("out_file",              "inputnode.in_file")]),
                    (bet_dwi0,  hmc, [("mask_file",             "inputnode.in_mask")]),
                    (list_b0,   hmc, [(("out_idx", pick_first), "inputnode.ref_num"),]),

                    # artifact detection
                    (hmc,      art, [("outputnode.out_file", "realigned_files"),
                                     ("outputnode.out_xfms", "realignment_parameters"),
                                    ]),
                    (bet_dwi1, art, [("mask_file", "mask_file"),]),

                    # output
                    (hmc, dti_output, [("outputnode.out_file", "hmc_corr_file"),
                                       ("outputnode.out_bvec", "hmc_corr_bvec"),
                                       ("outputnode.out_xfms", "hmc_corr_xfms"),
                                      ]),

                    (art, dti_output, [("displacement_files",  "art_displacement_files"),
                                       ("intensity_files",     "art_intensity_files"),
                                       ("norm_files",          "art_norm_files"),
                                       ("outlier_files",       "art_outlier_files"),
                                       ("plot_files",          "art_plot_files"),
                                       ("statistic_files",     "art_statistic_files"),
                                      ]),
                  ])

    return wf


def attach_dti_artifact_correction(main_wf, wf_name="dti_artifact_correction"):
    """ Attach the FSL-based diffusion MRI artifact detection and correction
    workflow to the `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow

    params: dict with parameter values
        atlas_file: str
            Path to the anatomical atlas to be transformed to diffusion MRI space.

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.select.diff: input node

    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)

    # The workflow box
    art_dti_wf = dti_artifact_correction(wf_name=wf_name)

    # dataSink output substitutions
    ## The base name of the 'diff' file for the substitutions
    diff_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'diff')))

    regexp_subst = [
                    (r"/brain_mask_{diff}_space\.nii$", "/brain_mask.nii"),
                    (r"/eddy_corrected\.nii$",          "/{diff}_eddycor.nii"),
                   ]
    regexp_subst = format_pair_list(regexp_subst, diff=diff_fbasename)

    regexp_subst += extension_duplicates(regexp_subst)
    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    # input and output diffusion MRI workflow to main workflow connections
    main_wf.connect([(in_files,   art_dti_wf, [("diff", "dti_art_input.diff"),
                                               ("bval", "dti_art_input.bval"),
                                               ("bvec", "dti_art_input.bvec"),
                                              ]),
                     (art_dti_wf, datasink,   [("dti_art_output.eddy_corr_file", "diff.@eddy_corr_file"),
                                               ("dti_art_output.bvec_rotated",   "diff.@bvec_rotated"),
                                               ("dti_art_output.brain_mask_1",   "diff.@brain_mask_1"),
                                               ("dti_art_output.brain_mask_2",   "diff.@brain_mask_2"),
                                               ("dti_art_output.acqp",           "diff.@acquisition_pars"),
                                               ("dti_art_output.index",          "diff.@acquisition_idx"),
                                               ("dti_art_output.avg_b0",         "diff.@avg_b0"),
                                              ]),
                    ])

    do_rapidart = get_config_setting("dmri.artifact_detect", True)
    if do_rapidart:
        main_wf.connect([(art_dti_wf, datasink, [
                                                 ("dti_art_output.hmc_corr_file",          "diff.artifact_stats.@hmc_corr_file"),
                                                 ("dti_art_output.hmc_corr_bvec",          "diff.artifact_stats.@hmc_rot_bvec"),
                                                 ("dti_art_output.hmc_corr_xfms",          "diff.artifact_stats.@hmc_corr_xfms"),
                                                 ("dti_art_output.art_displacement_files", "diff.artifact_stats.@art_disp_files"),
                                                 ("dti_art_output.art_intensity_files",    "diff.artifact_stats.@art_ints_files"),
                                                 ("dti_art_output.art_norm_files",         "diff.artifact_stats.@art_norm_files"),
                                                 ("dti_art_output.art_outlier_files",      "diff.artifact_stats.@art_outliers"),
                                                 ("dti_art_output.art_plot_files",         "diff.artifact_stats.@art_plots"),
                                                 ("dti_art_output.art_statistic_files",    "diff.artifact_stats.@art_stats"),
                                                ]),
                        ])

    return main_wf



# -*- coding: utf-8 -*-
"""
fMRI timeseries filtering helpers.
"""


def bandpass_filter(files, lowpass_freq=0.1, highpass_freq=0.01, tr=2):
    """Bandpass filter the input files

    Parameters
    ----------
    files: list of str
        List 4d nifti file paths.

    lowpass_freq: float
        Cutoff frequency for the low pass filter (in Hz).

    highpass_freq: float
        Cutoff frequency for the high pass filter (in Hz).

    tr: float
        The repetition time in seconds. The inverse of sampling rate (in Hz).
    """
    import os

    import nibabel as nb
    import numpy as np
    from   nipype.utils.filemanip import (filename_to_list,
                                          list_to_filename,
                                          split_filename)

    fs = 1./tr

    out_files = []
    for filename in filename_to_list(files):
        path, name, ext = split_filename(filename)
        out_file = os.path.join(os.getcwd(), name + '_bandpassed' + ext)

        img = nb.load(filename)
        timepoints = img.shape[-1]
        F = np.zeros((timepoints))

        lowidx = int(timepoints / 2) + 1
        if lowpass_freq > 0:
            lowidx = np.round(float(lowpass_freq) / fs * timepoints)

        highidx = 0
        if highpass_freq > 0:
            highidx = np.round(float(highpass_freq) / fs * timepoints)
        F[int(highidx):int(lowidx)] = 1
        F = ((F + F[::-1]) > 0).astype(int)
        data = img.get_data()
        if np.all(F == 1):
            filtered_data = data
        else:
            filtered_data = np.real(np.fft.ifftn(np.fft.fftn(data) * F))
        img_out = nb.Nifti1Image(filtered_data, img.affine, img.header)
        img_out.to_filename(out_file)
        out_files.append(out_file)

    return list_to_filename(out_files)



# -*- coding: utf-8 -*-
"""
Helper functions to plot nipype workflows
"""
from .interfaces.nilearn import (plot_multi_slices,
                                 plot_ortho_slices,
                                 plot_ica_components,
                                 plot_all_components)


def plot_workflow(wf, detailed=False):
    """ Plot the `wf` pipeline nodes in different ways in its working directory.

    Parameters
    ----------
    wf: nipype Workflow

    detailed: bool
        If True will also plot the detailed execution plot.
        This takes a very long time if the number of subjects if large.
    """
    # print the graph of the workflow in the working directory
    try:
        wf.write_graph("{}_colored_workflow".format(wf.name),
                       graph2use="colored")
    except RuntimeError as re:
        print('Error plotting colored workflow: {}.'.format(str(re)))
    else:
        if detailed:
            wf.write_graph("{}_exec_workflow".format(wf.name),
                           graph2use="exec")



# -*- coding: utf-8 -*-
"""
Nipype interface for ANTs.

The implementation of this interface is motivated by this study:
http://dx.doi.org/10.1016/j.nicl.2016.05.017
"""

from nipype.interfaces.base import TraitedSpec, File, traits, isdefined
from nipype.interfaces.ants.base import ANTSCommand, ANTSCommandInputSpec
from nipype.utils.filemanip import split_filename
from nipype.external.due import BibTeX


class KellyKapowskiInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(3, 2, argstr='--image-dimensionality %d', usedefault=True,
                            desc='image dimension (2 or 3)')

    segmentation_image = File(exists=True, argstr='--segmentation-image "%s"', mandatory=True,
                              desc="A segmentation image must be supplied labeling the gray and white matters.\n"
                                   "Default values = 2 and 3, respectively.",)

    gray_matter_label = traits.Int(2, usedefault=True,
                                   desc="The label value for the gray matter label in the segmentation_image.")

    white_matter_label = traits.Int(3, usedefault=True,
                                    desc="The label value for the white matter label in the segmentation_image.")

    gray_matter_prob_image = File(exists=True, argstr='--gray-matter-probability-image "%s"',
                                  desc="In addition to the segmentation image, a gray matter probability image can be\n"
                                       "used. If no such image is supplied, one is created using the segmentation image\n"
                                       "and a variance of 1.0 mm.")

    white_matter_prob_image = File(exists=True, argstr='--white-matter-probability-image "%s"',
                                   desc="In addition to the segmentation image, a white matter probability image can be\n"
                                       "used. If no such image is supplied, one is created using the segmentation image\n"
                                       "and a variance of 1.0 mm.")

    convergence = traits.Str(default="[50,0.001,10]", argstr='--convergence "%s"', usedefault=True,
                             desc="Convergence is determined by fitting a line to the normalized energy profile of\n"
                                  "the last N iterations (where N is specified by the window size) and determining\n"
                                  "the slope which is then compared with the convergence threshold.",)

    thickness_prior_estimate = traits.Float(10, usedefault=True, argstr="--thickness-prior-estimate %f",
                                            desc="Provides a prior constraint on the final thickness measurement in mm.")

    thickness_prior_image = File(exists=True, argstr='--thickness-prior-image "%s"',
                                 desc="An image containing spatially varying prior thickness values.")

    gradient_step = traits.Float(0.025, usedefault=True, argstr="--gradient-step %f",
                                 desc="Gradient step size for the optimization.")

    smoothing_variance = traits.Float(1.0, argstr="--smoothing-variance %f",
                                      desc="Defines the Gaussian smoothing of the hit and total images.")

    smoothing_velocity_field = traits.Float(1.5, argstr="--smoothing-velocity-field-parameter %f",
                                            desc="Defines the Gaussian smoothing of the velocity field (default = 1.5).\n"
                                            "If the b-spline smoothing option is chosen, then this defines the \n"
                                            "isotropic mesh spacing for the smoothing spline (default = 15).")

    use_bspline_smoothing = traits.Bool(argstr="--use-bspline-smoothing 1",
                                        desc="Sets the option for B-spline smoothing of the velocity field.")

    number_integration_points = traits.Int(10, argstr="--number-of-integration-points %d",
                                           desc="Number of compositions of the diffeomorphism per iteration.")

    max_invert_displacement_field_iters = traits.Int(20, argstr="--maximum-number-of-invert-displacement-field-iterations %d",
                                                     desc="Maximum number of iterations for estimating the invert \n"
                                                          "displacement field.")

    cortical_thickness = File(argstr='--output "%s"', keep_extension=True,
                              name_source=["segmentation_image"], name_template='%s_cortical_thickness',
                              desc='Filename for the cortical thickness.', hash_files=False)

    warped_white_matter = File(name_source=["segmentation_image"], keep_extension=True,
                               name_template='%s_warped_white_matter',
                               desc='Filename for the warped white matter file.', hash_files=False)


class KellyKapowskiOutputSpec(TraitedSpec):
    cortical_thickness = File(desc="A thickness map defined in the segmented gray matter.")
    warped_white_matter = File(desc="A warped white matter image.")


class KellyKapowski(ANTSCommand):
    """ Nipype Interface to ANTs' KellyKapowski, also known as DiReCT.

    DiReCT is a registration based estimate of cortical thickness. It was published
    in S. R. Das, B. B. Avants, M. Grossman, and J. C. Gee, Registration based
    cortical thickness measurement, Neuroimage 2009, 45:867--879.

    Examples
    --------
    >>> from nipype.interfaces.ants.segmentation import KellyKapowski
    >>> kk = KellyKapowski()
    >>> kk.inputs.dimension = 3
    >>> kk.inputs.segmentation_image = "segmentation0.nii.gz"
    >>> kk.inputs.convergence = "[45,0.0,10]"
    >>> kk.inputs.gradient_step = 0.025
    >>> kk.inputs.smoothing_variance = 1.0
    >>> kk.inputs.smoothing_velocity_field = 1.5
    >>> #kk.inputs.use_bspline_smoothing = False
    >>> kk.inputs.number_integration_points = 10
    >>> kk.inputs.thickness_prior_estimate = 10
    >>> kk.cmdline # doctest: +ALLOW_UNICODE
    u'KellyKapowski --convergence "[45,0.0,10]" \
--output "[segmentation0_cortical_thickness.nii.gz,segmentation0_warped_white_matter.nii.gz]" \
--image-dimensionality 3 --gradient-step 0.025000 --number-of-integration-points 10 \
--segmentation-image "[segmentation0.nii.gz,2,3]" --smoothing-variance 1.000000 \
--smoothing-velocity-field-parameter 1.500000 --thickness-prior-estimate 10.000000'

    """
    _cmd = "KellyKapowski"
    input_spec = KellyKapowskiInputSpec
    output_spec = KellyKapowskiOutputSpec

    references_ = [{'entry': BibTeX("@book{Das2009867,"
                                    "author={Sandhitsu R. Das and Brian B. Avants and Murray Grossman and James C. Gee},"
                                    "title={Registration based cortical thickness measurement.},"
                                    "journal={NeuroImage},"
                                    "volume={45},"
                                    "number={37},"
                                    "pages={867--879},"
                                    "year={2009},"
                                    "issn={1053-8119},"
                                    "url={http://www.sciencedirect.com/science/article/pii/S1053811908012780},"
                                    "doi={http://dx.doi.org/10.1016/j.neuroimage.2008.12.016}"
                                    "}"),
                    'description': 'The details on the implementation of DiReCT.',
                    'tags': ['implementation'],
                    }]

    def _parse_inputs(self, skip=None):
        if skip is None:
            skip = []
        skip += ['warped_white_matter', 'gray_matter_label', 'white_matter_label']
        return super(KellyKapowski, self)._parse_inputs(skip=skip)

    def _gen_filename(self, name):
        if name == 'cortical_thickness':
            output = self.inputs.cortical_thickness
            if not isdefined(output):
                _, name, ext = split_filename(self.inputs.segmentation_image)
                output = name + '_cortical_thickness' + ext
            return output

        if name == 'warped_white_matter':
            output = self.inputs.warped_white_matter
            if not isdefined(output):
                _, name, ext = split_filename(self.inputs.segmentation_image)
                output = name + '_warped_white_matter' + ext
            return output

        return None

    def _format_arg(self, opt, spec, val):
        if opt == "segmentation_image":
            newval = '[{0},{1},{2}]'.format(self.inputs.segmentation_image,
                                            self.inputs.gray_matter_label,
                                            self.inputs.white_matter_label)
            return spec.argstr % newval

        if opt == "cortical_thickness":
            ct = self._gen_filename("cortical_thickness")
            wm = self._gen_filename("warped_white_matter")
            newval = '[{},{}]'.format(ct, wm)
            return spec.argstr % newval

        return super(KellyKapowski, self)._format_arg(opt, spec, val)



# -*- coding: utf-8 -*-
"""
Helper functions to build base workflow and run them
"""
from neuro_pypes.plot import plot_workflow


def run_wf(wf, plugin='MultiProc', n_cpus=2, **plugin_kwargs):
    """ Execute `wf` with `plugin`.

    Parameters
    ----------
    wf: nipype Workflow

    plugin: str
        The pipeline execution plugin.
        See wf.run docstring for choices.

    n_cpus: int
        Number of CPUs to use with the 'MultiProc' plugin.

    plugin_kwargs: keyword argumens
        Keyword arguments for the plugin if using something different
        then 'MultiProc'.
    """
    # run the workflow according to `plugin`
    if plugin == "MultiProc" or n_cpus > 1:
        wf.run("MultiProc", plugin_args={"n_procs": n_cpus})
    elif not plugin or plugin is None or n_cpus <= 1:
        wf.run(plugin=None)
    else:
        wf.run(plugin=plugin, **plugin_kwargs)


def run_debug(workflow, plugin="MultiProc", n_cpus=4, **plugin_kwargs):
    """ Execute `wf` with `plugin`.

    Parameters
    ----------
    wf: nipype Workflow

    plugin: str
        The pipeline execution plugin.
        See wf.run docstring for choices.

    n_cpus: int
        Number of CPUs to use with the 'MultiProc' plugin.

    plugin_kwargs: keyword argumens
        Keyword arguments for the plugin if using something different
        then 'MultiProc'.
    """

    try:
        # plot the graph in its working directory
        plot_workflow(workflow)

        # run it
        run_wf(workflow, plugin=plugin, n_cpus=n_cpus,
               **plugin_kwargs)
    except:
        import sys
        print(sys.exc_info())

        import pdb
        pdb.post_mortem(sys.exc_info()[2])

        raise
    else:
        print('Workflow successfully finished.')



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
 Helper functions and nipype interface for
 reading DICOM files, specially from Siemens acquisitions
"""


def dcm_ascii_hdr(dcm_file):
    """ Return the CSA ASCII header from a Siemens DICOM file.

    References
    ----------
    http://www.nmr.mgh.harvard.edu/~greve/dicom-unpack
    https://groups.google.com/forum/#!msg/pydicom/-CmY-yK8Y7A/XaR4cUpizvgJ

    Note
    ----
    This only works for Siemens DICOM files.
    """
    def get_csa_header(dcm_file):
        import pydicom
        from nibabel.nicom import csareader as csar
        dcm_data = pydicom.read_file(dcm_file)
        return csar.get_csa_header(dcm_data, 'series')

    csa = get_csa_header(dcm_file)
    return csa['tags']['MrPhoenixProtocol']['items'][0]


def split_dcm_ahdr(ahdr):
    """ Return the protocol meta data and the ascconv part separately
    in list of lines.

    Parameters
    ----------
    ahdr: str

    Returns
    -------
    meta: list of str

    ascconv: list of str

    References
    ----------
    https://groups.google.com/forum/#!msg/pydicom/-CmY-yK8Y7A/XaR4cUpizvgJ

    Note
    ----
    This only works for Siemens DICOM files.
    """
    parts = ahdr.split('### ASCCONV BEGIN ###')

    meta   = parts[0].split('\n')
    ascconv = parts[1].split('### ASCCONV END ###')[0].split('\n')

    return meta, ascconv


# -*- coding: utf-8 -*-
"""
Nipype processing nodes for ROI images
"""
import os.path as op


from nipype.interfaces.base import BaseInterface, \
    BaseInterfaceInputSpec, traits, File, TraitedSpec
from nipype.utils.filemanip import split_filename

import nibabel as nib
from   boyle.nifti.roi import drain_rois

from ..utils.files import niftiimg_out


@niftiimg_out
def drain_rois_img(img):
    return drain_rois(img)


class DrainROIsInputSpec(BaseInterfaceInputSpec):
    img = File(exists=True, desc='volume with ROIS to be drained', mandatory=True)


class DrainROIsOutputSpec(TraitedSpec):
    out = File(exists=True, desc="the volume with the drained ROIs")


class DrainROIs(BaseInterface):
    input_spec  = DrainROIsInputSpec
    output_spec = DrainROIsOutputSpec

    def _run_interface(self, runtime):
        fname = self.inputs.img
        nu_img = drain_rois_img(fname)

        _, base, _ = split_filename(fname)
        nib.save(nu_img, base + '_drained.nii.gz')

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        fname = self.inputs.img
        _, base, _ = split_filename(fname)
        outputs["out"] = op.abspath(base + '_drained.nii.gz')
        return outputs



# -*- coding: utf-8 -*-
"""
Resting-state fMRI specific nuisance correction filtering workflow.
"""

import nipype.pipeline.engine       as pe
from   nipype.algorithms.rapidart   import ArtifactDetect
from   nipype.interfaces.utility    import Function, IdentityInterface, Merge
from   nipype.algorithms.confounds  import TSNR
from   nipype.interfaces            import fsl

from   ..config  import setup_node, _get_params_for
from   ..utils   import selectindex, rename
from   ..preproc import motion_regressors, extract_noise_components, create_regressors


def rapidart_fmri_artifact_detection():
    art = ArtifactDetect()
    art.inputs.use_differences      = [True, False]
    art.inputs.use_norm             = True
    art.inputs.zintensity_threshold = 2
    art.inputs.norm_threshold       = 1
    art.inputs.mask_type            = 'file'
    art.inputs.parameter_source     = 'NiPy'
    return art


def rest_noise_filter_wf(wf_name='rest_noise_removal'):
    """ Create a resting-state fMRI noise removal node.

    Nipype Inputs
    -------------
    rest_noise_input.in_file

    rest_noise_input.brain_mask

    rest_noise_input.wm_mask

    rest_noise_input.csf_mask

    rest_noise_input.motion_params
        Nipy motion parameters.

    Nipype Outputs
    --------------
    rest_noise_output.tsnr_file
        A SNR estimation volume file for QA purposes.

    rest_noise_output.motion_corrected
        The fMRI motion corrected image.

    rest_noise_output.nuis_corrected
        The resulting nuisance corrected image.
        This will be the same as 'motion_corrected' if compcor
        is disabled.

    rest_noise_output.motion_regressors
        Motion regressors file.

    rest_noise_output.compcor_regressors
        CompCor regressors file.

    rest_noise_output.art_displacement_files
        One image file containing the voxel-displacement timeseries.

    rest_noise_output.art_intensity_files
        One file containing the global intensity values determined
        from the brainmask.

    rest_noise_output.art_norm_files
        One file containing the composite norm.

    rest_noise_output.art_outlier_files
         One file containing a list of 0-based indices corresponding
         to outlier volumes.

    rest_noise_output.art_plot_files
        One image file containing the detected outliers.

    rest_noise_output.art_statistic_files
        One file containing information about the different types of
        artifacts and if design info is provided then details of
        stimulus correlated motion and a listing or artifacts by
        event type.

    Returns
    -------
    rm_nuisance_wf: nipype Workflow
    """

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    in_fields = ["in_file",
                 "brain_mask",
                 "wm_mask",
                 "csf_mask",
                 "motion_params",]

    out_fields = ["tsnr_file",
                  "motion_corrected",
                  "nuis_corrected",
                  "motion_regressors",
                  "compcor_regressors",
                  "gsr_regressors",
                  "art_displacement_files",
                  "art_intensity_files",
                  "art_norm_files",
                  "art_outlier_files",
                  "art_plot_files",
                  "art_statistic_files",
                 ]

    # input identities
    rest_noise_input = setup_node(IdentityInterface(fields=in_fields,
                                                    mandatory_inputs=True),
                                  name="rest_noise_input")

    # get the settings for filters
    filters = _get_params_for('rest_filter')

    # Compute TSNR on realigned data regressing polynomial up to order 2
    tsnr = setup_node(TSNR(regress_poly=2), name='tsnr')

    # Use :class:`nipype.algorithms.rapidart` to determine which of the
    # images in the functional series are outliers based on deviations in
    # intensity or movement.
    art = setup_node(rapidart_fmri_artifact_detection(), name="detect_artifacts")

    # Compute motion regressors
    motion_regs = setup_node(Function(input_names=['motion_params',
                                                   'order',
                                                   'derivatives',
                                                  ],
                                      output_names=['out_files'],
                                      function=motion_regressors,),
                             name='motion_regressors')

    # Create a filter to remove motion and art confounds
    motart_pars = setup_node(Function(input_names=['motion_params',
                                                   'comp_norm',
                                                   'outliers',
                                                   'detrend_poly'],
                                      output_names=['out_files'],
                                      function=create_regressors),
                             name='motart_parameters')

    motion_filter = setup_node(fsl.GLM(out_f_name='F_mcart.nii.gz',
                                       out_pf_name='pF_mcart.nii.gz',
                                       demean=True),
                               name='motion_filter')

    # Noise confound regressors
    compcor_pars = setup_node(Function(input_names=['realigned_file',
                                                    'mask_file',
                                                    'num_components',
                                                    'extra_regressors'],
                                         output_names=['out_files'],
                                         function=extract_noise_components,),
                              name='compcor_pars')
    #compcor_pars = setup_node(ACompCor(), name='compcor_pars')
    #compcor_pars.inputs.components_file = 'noise_components.txt'

    compcor_filter = setup_node(fsl.GLM(out_f_name='F.nii.gz',
                                        out_pf_name='pF.nii.gz',
                                        demean=True),
                                name='compcor_filter')

    # Global signal regression
    gsr_pars = setup_node(Function(input_names=['realigned_file',
                                                'mask_file',
                                                'num_components',
                                                'extra_regressors'],
                                    output_names=['out_files'],
                                    function=extract_noise_components, ),
                            name='gsr_pars')

    gsr_filter = setup_node(fsl.GLM(out_f_name='F_gsr.nii.gz',
                                    out_pf_name='pF_gsr.nii.gz',
                                    demean=True),
                            name='gsr_filter')

    # output identities
    rest_noise_output = setup_node(IdentityInterface(fields=out_fields,
                                                     mandatory_inputs=True),
                                    name="rest_noise_output")

    # Connect the nodes
    wf.connect([
                # tsnr
                (rest_noise_input, tsnr, [("in_file", "in_file")]),

                # artifact detection
                (rest_noise_input, art, [("in_file",       "realigned_files"),
                                         ("motion_params", "realignment_parameters"),
                                         ("brain_mask",    "mask_file"),
                                        ]),

                # calculte motion regressors
                (rest_noise_input,  motion_regs, [("motion_params", "motion_params")]),

                # create motion and confound regressors parameters file
                (art,           motart_pars, [("norm_files",    "comp_norm"),
                                              ("outlier_files", "outliers"),
                                             ]),
                (motion_regs,   motart_pars, [("out_files", "motion_params")]),

                # motion filtering
                (rest_noise_input, motion_filter, [("in_file", "in_file"),
                                                   (("in_file", rename, "_filtermotart"), "out_res_name"),
                                                  ]),
                (motart_pars,      motion_filter, [(("out_files", selectindex, 0), "design")]),

                # output
                (tsnr,             rest_noise_output, [("tsnr_file",          "tsnr_file")]),
                (motart_pars,      rest_noise_output, [("out_files",          "motion_regressors")]),
                (motion_filter,    rest_noise_output, [("out_res",            "motion_corrected")]),
                (art,              rest_noise_output, [("displacement_files", "art_displacement_files"),
                                                       ("intensity_files",    "art_intensity_files"),
                                                       ("norm_files",         "art_norm_files"),
                                                       ("outlier_files",      "art_outlier_files"),
                                                       ("plot_files",         "art_plot_files"),
                                                       ("statistic_files",    "art_statistic_files"),
                                                      ]),
                ])

    last_filter = motion_filter

    # compcor filter
    if filters['compcor_csf'] or filters['compcor_wm']:
        wf.connect([
                    # calculate compcor regressor and parameters file
                    (motart_pars,      compcor_pars,      [(("out_files", selectindex, 0),   "extra_regressors"),]),
                    (motion_filter,    compcor_pars,      [("out_res",                       "realigned_file"),]),

                    # the compcor filter
                    (motion_filter,    compcor_filter,    [("out_res",                        "in_file"),
                                                           (("out_res", rename, "_cleaned"),  "out_res_name"),
                                                          ]),
                    (compcor_pars,     compcor_filter,    [(("out_files", selectindex, 0),    "design")]),
                    #(compcor_pars,     compcor_filter,    [("components_file",  "design")]),
                    (rest_noise_input, compcor_filter,    [("brain_mask",                     "mask")]),

                    # output
                    (compcor_pars,     rest_noise_output, [("out_files",   "compcor_regressors")]),
                    #(compcor_pars,     rest_noise_output, [("components_file",   "compcor_regressors")]),
                    ])
        last_filter = compcor_filter

    # global signal regression
    if filters['gsr']:
        wf.connect([
            # calculate gsr regressors parameters file
            (last_filter,       gsr_pars, [("out_res",      "realigned_file")]),
            (rest_noise_input,  gsr_pars, [("brain_mask",   "mask_file")]),

            # the output file name
            (rest_noise_input, gsr_filter,  [("brain_mask", "mask")]),
            (last_filter,       gsr_filter, [("out_res", "in_file"),
                                             (("out_res", rename, "_gsr"), "out_res_name"),
                                            ]),
            (gsr_pars,          gsr_filter, [(("out_files", selectindex, 0), "design")]),

            # output
            (gsr_pars,   rest_noise_output, [("out_files", "gsr_regressors")]),
        ])
        last_filter = gsr_filter

    # connect the final nuisance correction output node
    wf.connect([(last_filter, rest_noise_output, [("out_res", "nuis_corrected")]),])

    if filters['compcor_csf'] and filters['compcor_wm']:
        mask_merge = setup_node(Merge(2), name="mask_merge")
        wf.connect([
                    ## the mask for the compcor filter
                    (rest_noise_input, mask_merge,   [("wm_mask",    "in1")]),
                    (rest_noise_input, mask_merge,   [("csf_mask",   "in2")]),
                    (mask_merge,       compcor_pars, [("out",        "mask_file")]),
                  ])

    elif filters['compcor_csf']:
        wf.connect([
                    ## the mask for the compcor filter
                    (rest_noise_input, compcor_pars, [("csf_mask",   "mask_file")]),
                  ])

    elif filters['compcor_wm']:
        wf.connect([
                    ## the mask for the compcor filter
                    (rest_noise_input, compcor_pars, [("wm_mask",   "mask_file")]),
                  ])

    return wf



# -*- coding: utf-8 -*-
"""
Nipype utilities for processing anatomical MRI.
"""
import nipype.interfaces.spm as spm
from   nipype.interfaces.ants import N4BiasFieldCorrection
from   nipype.interfaces.base import traits

from   ..utils import spm_tpm_priors_path


def biasfield_correct(anat_filepath=traits.Undefined):
    """ Inhomogeneity correction.
    ANTS N4BiasFieldCorrection interface.

    Parameters
    ----------
    anat_filepath: str
        Path to the anatomical file path

    Returns
    -------
    seg: N4BiasFieldCorrection interface
    """
    n4 = N4BiasFieldCorrection()
    n4.inputs.dimension = 3
    n4.inputs.bspline_fitting_distance = 300
    n4.inputs.shrink_factor = 3
    n4.inputs.n_iterations = [50, 50, 30, 20]
    n4.inputs.convergence_threshold = 1e-6
    #n4.inputs.bspline_order = 5
    n4.inputs.save_bias = True

    n4.inputs.input_image = anat_filepath

    return n4


def spm_segment(anat_filepath=traits.Undefined, priors_path=None):
    """ SPM12 New Segment interface.

    Parameters
    ----------
    anat_filepath: str
        Path to the anatomical file path

    priors_path: str
        Path to the tissue probability maps file

    Returns
    -------
    seg: NewSegment interface
    """
    if priors_path is None:
        priors_path = spm_tpm_priors_path()

    seg = spm.NewSegment()

    tissue1 = ((priors_path, 1), 1, (True,  True),   (True,  True))
    tissue2 = ((priors_path, 2), 1, (True,  True),   (True,  True))
    tissue3 = ((priors_path, 3), 2, (True,  True),   (True,  True))
    tissue4 = ((priors_path, 4), 3, (True,  True),   (True,  True))
    tissue5 = ((priors_path, 5), 4, (True,  False),  (False, False))
    tissue6 = ((priors_path, 6), 2, (False, False),  (False, False))
    seg.inputs.tissues = [tissue1, tissue2, tissue3, tissue4, tissue5, tissue6]
    seg.inputs.channel_info = (0.0001, 60, (True, True))
    #seg.inputs.warping_regularization = [0, 0.001, 0.5, 0.05, 0.2]
    seg.inputs.write_deformation_fields = [True, True]

    seg.inputs.channel_files = anat_filepath

    #seg.run()
    return seg



# -*- coding: utf-8 -*-

from .petpvc import PETPVC
from .realign import nipy_motion_correction
from .denoise import (nlmeans_denoise_img,
                      create_regressors,
                      extract_noise_components,
                      motion_regressors,
                      reslice_img)
from .registration import (spm_apply_deformations,
                           spm_coregister,
                           spm_normalize,
                           spm_warp_to_mni,
                           afni_deoblique,
                           spm_tpm_priors_path,
                           spm_create_group_template_wf,
                           spm_register_to_template_wf,
                           )
from .slicetime import (afni_slicetime,
                        spm_slicetime,
                        auto_spm_slicetime,
                        auto_nipy_slicetime)
from .slicetime_params import (STCParameters,
                               STCParametersInterface)

from .spatial import get_bounding_box



# -*- coding: utf-8 -*-
"""
fMRI pre-processing workflows
"""

import nipype.pipeline.engine as pe
import nipype.interfaces.fsl as fsl
import nipype.interfaces as util
from   nipype.interfaces.afni import preprocess


# workflow to edit the scan to the proscribed TRs
def create_wf_edit_func(wf_name = "edit_func" ):
    """
    Workflow Inputs::
    
        inputspec.func : func file or a list of func/rest nifti file 
            User input functional(T2*) Image
            
        inputspec.start_idx : string 
            Starting volume/slice of the functional image (optional)
            
        inputspec.stop_idx : string
            Last volume/slice of the functional image (optional)
            
    Workflow Outputs::
    
        outputspec.edited_func : string (nifti file)
            Path to Output image with the initial few slices dropped
          
           
    Order of commands:
    
    - Get the start and the end volume index of the functional run. If not defined by the user, return the first and last volume.
    
        get_idx(in_files, stop_idx, start_idx)
        
    - Dropping the initial TRs. For details see `3dcalc <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3dcalc.html>`_::
        
        3dcalc -a rest.nii.gz[4..299] 
               -expr 'a' 
               -prefix rest_3dc.nii.gz
               
    """

    # allocate a workflow object
    try:
        preproc = pe.Workflow(name=wf_name)
    except:
        logger.info( "Error allocating workflow %s."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

    # configure the workflow's input spec
    try:
        inputNode = pe.Node(util.IdentityInterface(fields=['func',
                                                           'start_idx',
                                                           'stop_idx']),
                            name='inputspec')
    except:
        logger.info( "Error allocating inputspec (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

    # configure the workflow's output spec
    try:
        outputNode = pe.Node(util.IdentityInterface(fields=['edited_func']),
                          name='outputspec')
    except:
        logger.info( "Error allocating output spec (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

    # allocate a node to check that the requested edits are
    # reasonable given the data
    try:
        func_get_idx = pe.Node(util.Function(input_names=['in_files', 
                                                          'stop_idx', 
                                                          'start_idx'],
                                   output_names=['stopidx', 
                                                 'startidx'],
                     function=get_idx), name='func_get_idx')
    except:
        logger.info( "Error allocating get_idx function node (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

   
    # wire in the func_get_idx node
    try: 
        preproc.connect(inputNode, 'func',
                        func_get_idx, 'in_files')
    except:
        logger.info( "Error connecting 'in_files' input to get_idx function node (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

    try:
        preproc.connect(inputNode, 'start_idx',
                        func_get_idx, 'start_idx')
    except:
        logger.info( "Error connecting 'start_idx' input to get_idx function node (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

    try:
        preproc.connect(inputNode, 'stop_idx',
                     func_get_idx, 'stop_idx')
    except:
        logger.info( "Error connecting 'stop_idx' input to get_idx function node (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

    try:
        # allocate a node to edit the functional file 
        func_drop_trs = pe.Node(interface=preprocess.Calc(),
                               name='func_drop_trs')
        func_drop_trs.inputs.expr = 'a'
        func_drop_trs.inputs.outputtype = 'NIFTI_GZ'
    except:
        logger.info( "Error allocating afni Calc node (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

   
    # wire in the inpus
    try: 
        preproc.connect(inputNode, 'func',
                        func_drop_trs, 'in_file_a')
    except:
        logger.info( "Error connecting 'in_file_a' input to afni Calc node (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

    try: 
        preproc.connect(func_get_idx, 'startidx',
                        func_drop_trs, 'start_idx')
    except:
        logger.info( "Error connecting 'start_idx' input to afni Calc node (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise

    try: 
        preproc.connect(func_get_idx, 'stopidx',
                        func_drop_trs, 'stop_idx')
    except:
        logger.info( "Error connecting 'stop_idx' input to afni Calc node (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise
  
    try: 
        # wire the output 
        preproc.connect(func_drop_trs, 'out_file',
                        outputNode, 'edited_func')
    except:
        logger.info( "Error connecting output (wflow %s)."+\
                     " (%s:%d)" % (wf_name, dbg_file_lineno() ))
        raise
  
    return preproc
    


#functional preprocessing
def create_func_preproc(use_bet = False, wf_name = 'func_preproc'):
    """
    
    The main purpose of this workflow is to process functional data. Raw rest file is deobliqued and reoriented 
    into RPI. Then take the mean intensity values over all time points for each voxel and use this image 
    to calculate motion parameters. The image is then skullstripped, normalized and a processed mask is 
    obtained to use it further in Image analysis.
    
    Parameters
    ----------
    
    wf_name : string
        Workflow name
    
    Returns 
    -------
    func_preproc : workflow object
        Functional Preprocessing workflow object
    
    Notes
    -----
    
    `Source <https://github.com/FCP-INDI/C-PAC/blob/master/CPAC/func_preproc/func_preproc.py>`_
    
    Workflow Inputs::
    
        inputspec.rest : func/rest file or a list of func/rest nifti file 
            User input functional(T2) Image, in any of the 8 orientations
            
        scan_params.tr : string
            Subject TR
        
        scan_params.acquistion : string
            Acquisition pattern (interleaved/sequential, ascending/descending)
    
        scan_params.ref_slice : integer
            Reference slice for slice timing correction
            
    Workflow Outputs::
    
        outputspec.refit : string (nifti file)
            Path to deobliqued anatomical data 
        
        outputspec.reorient : string (nifti file)
            Path to RPI oriented anatomical data 
        
        outputspec.motion_correct_ref : string (nifti file)
             Path to Mean intensity Motion corrected image 
             (base reference image for the second motion correction run)
        
        outputspec.motion_correct : string (nifti file)
            Path to motion corrected output file
        
        outputspec.max_displacement : string (Mat file)
            Path to maximum displacement (in mm) for brain voxels in each volume
        
        outputspec.movement_parameters : string (Mat file)
            Path to 1D file containing six movement/motion parameters(3 Translation, 3 Rotations) 
            in different columns (roll pitch yaw dS  dL  dP)
        
        outputspec.skullstrip : string (nifti file)
            Path to skull stripped Motion Corrected Image 
        
        outputspec.mask : string (nifti file)
            Path to brain-only mask
            
        outputspec.example_func : string (nifti file)
            Mean, Skull Stripped, Motion Corrected output T2 Image path
            (Image with mean intensity values across voxels) 
        
        outputpsec.preprocessed : string (nifti file)
            output skull stripped, motion corrected T2 image 
            with normalized intensity values 

        outputspec.preprocessed_mask : string (nifti file)
           Mask obtained from normalized preprocessed image
           
    Order of commands:
               
    - Deobliqing the scans.  For details see `3drefit <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3drefit.html>`_::
    
        3drefit -deoblique rest_3dc.nii.gz
        
    - Re-orienting the Image into Right-to-Left Posterior-to-Anterior Inferior-to-Superior (RPI) orientation. For details see `3dresample <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3dresample.html>`_::
    
        3dresample -orient RPI 
                   -prefix rest_3dc_RPI.nii.gz 
                   -inset rest_3dc.nii.gz
        
    - Calculate voxel wise statistics. Get the RPI Image with mean intensity values over all timepoints for each voxel. For details see `3dTstat <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3dTstat.html>`_::
    
        3dTstat -mean 
                -prefix rest_3dc_RPI_3dT.nii.gz 
                rest_3dc_RPI.nii.gz
    
    - Motion Correction. For details see `3dvolreg <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3dvolreg.html>`_::  
       
        3dvolreg -Fourier 
                 -twopass 
                 -base rest_3dc_RPI_3dT.nii.gz/
                 -zpad 4 
                 -maxdisp1D rest_3dc_RPI_3dvmd1D.1D 
                 -1Dfile rest_3dc_RPI_3dv1D.1D 
                 -prefix rest_3dc_RPI_3dv.nii.gz 
                 rest_3dc_RPI.nii.gz
                 
      The base image or the reference image is the mean intensity RPI image obtained in the above the step.For each volume 
      in RPI-oriented T2 image, the command, aligns the image with the base mean image and calculates the motion, displacement 
      and movement parameters. It also outputs the aligned 4D volume and movement and displacement parameters for each volume.
                 
    - Calculate voxel wise statistics. Get the motion corrected output Image from the above step, with mean intensity values over all timepoints for each voxel. 
      For details see `3dTstat <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3dTstat.html>`_::
    
        3dTstat -mean 
                -prefix rest_3dc_RPI_3dv_3dT.nii.gz 
                rest_3dc_RPI_3dv.nii.gz
    
    - Motion Correction and get motion, movement and displacement parameters. For details see `3dvolreg <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3dvolreg.html>`_::   

        3dvolreg -Fourier 
                 -twopass 
                 -base rest_3dc_RPI_3dv_3dT.nii.gz 
                 -zpad 4 
                 -maxdisp1D rest_3dc_RPI_3dvmd1D.1D 
                 -1Dfile rest_3dc_RPI_3dv1D.1D 
                 -prefix rest_3dc_RPI_3dv.nii.gz 
                 rest_3dc_RPI.nii.gz
        
      The base image or the reference image is the mean intensity motion corrected image obtained from the above the step (first 3dvolreg run). 
      For each volume in RPI-oriented T2 image, the command, aligns the image with the base mean image and calculates the motion, displacement 
      and movement parameters. It also outputs the aligned 4D volume and movement and displacement parameters for each volume.
    
    - Create a  brain-only mask. For details see `3dautomask <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3dAutomask.html>`_::
    
        3dAutomask  
                   -prefix rest_3dc_RPI_3dv_automask.nii.gz 
                   rest_3dc_RPI_3dv.nii.gz

    - Edge Detect(remove skull) and get the brain only. For details see `3dcalc <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3dcalc.html>`_::
    
        3dcalc -a rest_3dc_RPI_3dv.nii.gz 
               -b rest_3dc_RPI_3dv_automask.nii.gz 
               -expr 'a*b' 
               -prefix rest_3dc_RPI_3dv_3dc.nii.gz
    
    - Normalizing the image intensity values. For details see `fslmaths <http://www.fmrib.ox.ac.uk/fsl/avwutils/index.html>`_::
      
        fslmaths rest_3dc_RPI_3dv_3dc.nii.gz 
                 -ing 10000 rest_3dc_RPI_3dv_3dc_maths.nii.gz 
                 -odt float
                 
      Normalized intensity = (TrueValue*10000)/global4Dmean
                 
    - Calculate mean of skull stripped image. For details see `3dTstat <http://afni.nimh.nih.gov/pub/dist/doc/program_help/3dTstat.html>`_::
        
        3dTstat -mean -prefix rest_3dc_RPI_3dv_3dc_3dT.nii.gz rest_3dc_RPI_3dv_3dc.nii.gz
        
    - Create Mask (Generate mask from Normalized data). For details see `fslmaths <http://www.fmrib.ox.ac.uk/fsl/avwutils/index.html>`_::
        
        fslmaths rest_3dc_RPI_3dv_3dc_maths.nii.gz 
               -Tmin -bin rest_3dc_RPI_3dv_3dc_maths_maths.nii.gz 
               -odt char

    High Level Workflow Graph:
    
    .. image:: ../images/func_preproc.dot.png
       :width: 1000
    
    
    Detailed Workflow Graph:
    
    .. image:: ../images/func_preproc_detailed.dot.png
       :width: 1000

    Examples
    --------
    
    >>> import func_preproc
    >>> preproc = create_func_preproc(bet=True)
    >>> preproc.inputs.inputspec.func='sub1/func/rest.nii.gz'
    >>> preproc.run() #doctest: +SKIP


    >>> import func_preproc
    >>> preproc = create_func_preproc(bet=False)
    >>> preproc.inputs.inputspec.func='sub1/func/rest.nii.gz'
    >>> preproc.run() #doctest: +SKIP
    
    """

    preproc = pe.Workflow(name=wf_name)
    inputNode = pe.Node(util.IdentityInterface(fields=['func']),
                        name='inputspec')
    
    outputNode = pe.Node(util.IdentityInterface(fields=['refit',
                                                        'reorient',
                                                        'reorient_mean',
                                                        'motion_correct',
                                                        'motion_correct_ref',
                                                        'movement_parameters',
                                                        'max_displacement',
                                                        #'xform_matrix',
                                                        'mask',
                                                        'skullstrip',
                                                        'example_func',
                                                        'preprocessed',
                                                        'preprocessed_mask',
                                                        'slice_time_corrected',
                                                        'oned_matrix_save']),

                          name='outputspec')

    
    func_deoblique = pe.Node(interface=preprocess.Refit(),
                            name='func_deoblique')
    func_deoblique.inputs.deoblique = True
    
    preproc.connect(inputNode, 'func',
                        func_deoblique, 'in_file')
    
    func_reorient = pe.Node(interface=preprocess.Resample(),
                               name='func_reorient')
    func_reorient.inputs.orientation = 'RPI'
    func_reorient.inputs.outputtype = 'NIFTI_GZ'

    preproc.connect(func_deoblique, 'out_file',
                    func_reorient, 'in_file')
    
    preproc.connect(func_reorient, 'out_file',
                    outputNode, 'reorient')
    
    func_get_mean_RPI = pe.Node(interface=preprocess.TStat(),
                            name='func_get_mean_RPI')
    func_get_mean_RPI.inputs.options = '-mean'
    func_get_mean_RPI.inputs.outputtype = 'NIFTI_GZ'
    
    preproc.connect(func_reorient, 'out_file',
                    func_get_mean_RPI, 'in_file')
        
    #calculate motion parameters
    func_motion_correct = pe.Node(interface=preprocess.Volreg(),
                             name='func_motion_correct')
    func_motion_correct.inputs.args = '-Fourier -twopass'
    func_motion_correct.inputs.zpad = 4
    func_motion_correct.inputs.outputtype = 'NIFTI_GZ'

    
    preproc.connect(func_reorient, 'out_file',
                    func_motion_correct, 'in_file')
    preproc.connect(func_get_mean_RPI, 'out_file',
                    func_motion_correct, 'basefile')


    func_get_mean_motion = func_get_mean_RPI.clone('func_get_mean_motion')
    preproc.connect(func_motion_correct, 'out_file',
                    func_get_mean_motion, 'in_file')
    
    preproc.connect(func_get_mean_motion, 'out_file',
                    outputNode, 'motion_correct_ref')
    
    
    func_motion_correct_A = func_motion_correct.clone('func_motion_correct_A')
    func_motion_correct_A.inputs.md1d_file = 'max_displacement.1D'
    
    preproc.connect(func_reorient, 'out_file',
                    func_motion_correct_A, 'in_file')
    preproc.connect(func_get_mean_motion, 'out_file',
                    func_motion_correct_A, 'basefile')
    
    preproc.connect(func_motion_correct_A, 'out_file',
                    outputNode, 'motion_correct')
    preproc.connect(func_motion_correct_A, 'md1d_file',
                    outputNode, 'max_displacement')
    preproc.connect(func_motion_correct_A, 'oned_file',
                    outputNode, 'movement_parameters')
    preproc.connect(func_motion_correct_A, 'oned_matrix_save',
                    outputNode, 'oned_matrix_save')


    if use_bet == False:
    
        func_get_brain_mask = pe.Node(interface=preprocess.Automask(),
                                   name='func_get_brain_mask')

        func_get_brain_mask.inputs.outputtype = 'NIFTI_GZ'

        preproc.connect(func_motion_correct_A, 'out_file',
                        func_get_brain_mask, 'in_file')
    
        preproc.connect(func_get_brain_mask, 'out_file',
                        outputNode, 'mask')

    else:

        func_get_brain_mask = pe.Node(interface=fsl.BET(),
                                      name='func_get_brain_mask_BET')

        func_get_brain_mask.inputs.mask = True
        func_get_brain_mask.inputs.functional = True

        erode_one_voxel = pe.Node(interface=fsl.ErodeImage(),
                                  name='erode_one_voxel')

        erode_one_voxel.inputs.kernel_shape = 'box'
        erode_one_voxel.inputs.kernel_size = 1.0

        preproc.connect(func_motion_correct_A, 'out_file',
                        func_get_brain_mask, 'in_file')

        preproc.connect(func_get_brain_mask, 'mask_file',
                        erode_one_voxel, 'in_file')

        preproc.connect(erode_one_voxel, 'out_file',
                        outputNode, 'mask')

        
    
    func_edge_detect = pe.Node(interface=preprocess.Calc(),
                            name='func_edge_detect')
    func_edge_detect.inputs.expr = 'a*b'
    func_edge_detect.inputs.outputtype = 'NIFTI_GZ'

    preproc.connect(func_motion_correct_A, 'out_file',
                    func_edge_detect, 'in_file_a')


    if use_bet == False:

        preproc.connect(func_get_brain_mask, 'out_file',
                        func_edge_detect, 'in_file_b')

    else:

        preproc.connect(erode_one_voxel, 'out_file',
                        func_edge_detect, 'in_file_b')



    preproc.connect(func_edge_detect, 'out_file',
                    outputNode, 'skullstrip')

    
    func_mean_skullstrip = pe.Node(interface=preprocess.TStat(),
                           name='func_mean_skullstrip')
    func_mean_skullstrip.inputs.options = '-mean'
    func_mean_skullstrip.inputs.outputtype = 'NIFTI_GZ'

    preproc.connect(func_edge_detect, 'out_file',
                    func_mean_skullstrip, 'in_file')
    
    preproc.connect(func_mean_skullstrip, 'out_file',
                    outputNode, 'example_func')
    
    
    func_normalize = pe.Node(interface=fsl.ImageMaths(),
                            name='func_normalize')
    func_normalize.inputs.op_string = '-ing 10000'
    func_normalize.inputs.out_data_type = 'float'

    preproc.connect(func_edge_detect, 'out_file',
                    func_normalize, 'in_file')
    
    preproc.connect(func_normalize, 'out_file',
                    outputNode, 'preprocessed')
    
    
    func_mask_normalize = pe.Node(interface=fsl.ImageMaths(),
                           name='func_mask_normalize')
    func_mask_normalize.inputs.op_string = '-Tmin -bin'
    func_mask_normalize.inputs.out_data_type = 'char'

    preproc.connect(func_normalize, 'out_file',
                    func_mask_normalize, 'in_file')
    
    preproc.connect(func_mask_normalize, 'out_file',
                    outputNode, 'preprocessed_mask')

    return preproc



def get_idx(in_files, stop_idx=None, start_idx=None):

    """
    Method to get the first and the last slice for
    the functional run. It verifies the user specified
    first and last slice. If the values are not valid, it 
    calculates and returns the very first and the last slice 
    
    Parameters
    ----------
    in_file : string (nifti file)
       Path to input functional run
        
    stop_idx : int
        Last volume to be considered, specified by user
        in the configuration file 
    
    stop_idx : int
        First volume to be considered, specified by user 
        in the configuration file 
    
    Returns
    -------
    stop_idx :  int
        Value of first slice to consider for the functional run 
        
    start_idx : int 
        Value of last slice to consider for the functional run
        
    """

    #stopidx = None
    #startidx = None
    # Import packages
    from nibabel import load

    # Init variables
    img = load(in_files)
    hdr = img.get_header()
    shape = hdr.get_data_shape()
    
    # Check to make sure the input file is 4-dimensional
    if len(shape) != 4:
        raise TypeError('Input nifti file: %s is not a 4D file' % in_files)
    # Grab the number of volumes
    nvols = int(hdr.get_data_shape()[3])

    if (start_idx == None) or (start_idx < 0) or (start_idx > (nvols - 1)):
        startidx = 0
    else:
        startidx = start_idx

    if (stop_idx == None) or (stop_idx > (nvols - 1)):
        stopidx = nvols - 1
    else:
        stopidx = stop_idx

    return stopidx, startidx





from .config import PYPES_CFG as configuration

from .config import (get_config_setting,
                     check_mandatory_inputs,
                     node_settings,
                     setup_node,
                     _get_params_for,
                     check_atlas_file,
                     update_config)



# -*- coding: utf-8 -*-
"""
Denoise and motion correction helper functions
"""
from boyle.nifti.utils import nifti_out


@nifti_out
def nlmeans_denoise_img(img, mask, N=4):
    """ Apply dipy nlmeans denoising to the img. Useful for diffusion images.
    Parameters
    ----------
    img: nibabel.Nifti1Image
        The diffusion image

    mask: nibabel.Nifti1Image
        A brain mask image.

    N: int
        Number of arrays of the head coil used to acquired the image.

    Returns
    -------
    den_img: nibabel.Nifti1Image
        A denoised nifti image object with the same headers
        and affine as `img`.
    """
    from dipy.denoise.nlmeans import nlmeans
    from dipy.denoise.noise_estimate import estimate_sigma

    data = img.get_data()
    msk  = mask.get_data()

    sigma = estimate_sigma(data, N=N)
    return nlmeans(data, sigma=sigma, mask=msk)


def reslice_img(img, new_zooms=None, order=3):
    """ Performs regridding of an image to set isotropic voxel sizes using dipy.
    If the file has already isotropic voxels, will return a copy of the same image.

    Parameters
    ----------
    img: nibabel.Nifti1Image
        The diffusion image.

    new_zooms : tuple, shape (3,)
        new voxel size for (i,j,k) after resampling

    order : int, from 0 to 5
       order of interpolation for resampling/reslicing, 0 nearest interpolation, 1 trilinear etc..
       if you don’t want any smoothing 0 is the option you need.

    Returns
    -------
    nu_img: nibabel.Nifti1Image
        A isotropic voxel version from `img`.
    """
    import numpy as np
    import nibabel as nib
    from dipy.align.reslice import reslice

    # read the data
    data = img.get_data()
    img_zooms = img.header.get_zooms()[:3]

    # check if already isotropic voxels
    all_equal = len(np.unique(img_zooms)) == 1
    if all_equal:
        return nib.Nifti1Image(img.get_data(), affine=img.affine, header=img.header)

    # set new_zooms parameter
    if new_zooms is None:
        minzoom = np.array(img_zooms).min()
        new_zooms = tuple(np.ones((3,)) * minzoom)

    # reslice it
    nu_data, nu_affine = reslice(data=data, affine=img.affine,
                                 zooms=img_zooms,
                                 new_zooms=new_zooms,
                                 order=order)

    # create the new image object
    header = img.header.copy()
    tmp_zooms = np.array(header.get_zooms())
    tmp_zooms[:3] = new_zooms[0]
    header.set_zooms(tuple(tmp_zooms))
    header.set_data_shape(nu_data.shape)
    header.set_xyzt_units('mm')
    nu_img = nib.Nifti1Image(nu_data.astype(header.get_data_dtype()), nu_affine, header)

    return nu_img


def motion_regressors(motion_params, order=0, derivatives=1):
    """Compute motion regressors upto given order and derivative

    motion + d(motion)/dt + d2(motion)/dt2 (linear + quadratic)
    """
    import os
    import numpy as np
    from nipype.utils.filemanip import filename_to_list

    out_files = []
    for idx, filename in enumerate(filename_to_list(motion_params)):
        params = np.genfromtxt(filename)
        out_params = params
        for d in range(1, derivatives + 1):
            cparams = np.vstack((np.repeat(params[0, :][None, :], d, axis=0),
                                 params))
            out_params = np.hstack((out_params, np.diff(cparams, d, axis=0)))
        out_params2 = out_params
        for i in range(2, order + 1):
            out_params2 = np.hstack((out_params2, np.power(out_params, i)))
        filename = os.path.join(os.getcwd(), "motion_regressor%02d.txt" % idx)
        np.savetxt(filename, out_params2, fmt=b"%.10f")
        out_files.append(filename)
    return out_files


def create_regressors(motion_params, comp_norm, outliers, detrend_poly=None):
    """Builds a regressor set comprising motion parameters, composite norm and
    outliers.
    The outliers are added as a single time point column for each outlier

    Parameters
    ----------
    motion_params: a text file containing motion parameters and its derivatives
    comp_norm: a text file containing the composite norm
    outliers: a text file containing 0-based outlier indices
    detrend_poly: number of polynomials to add to detrend

    Returns
    -------
    components_file: a text file containing all the regressors
    """
    import os
    import numpy as np
    from nipype.utils.filemanip import filename_to_list
    from scipy.special import legendre

    out_files = []
    for idx, filename in enumerate(filename_to_list(motion_params)):
        params = np.genfromtxt(filename)
        norm_val = np.genfromtxt(filename_to_list(comp_norm)[idx])
        out_params = np.hstack((params, norm_val[:, None]))
        try:
            outlier_val = np.genfromtxt(filename_to_list(outliers)[idx])
        except IOError:
            outlier_val = np.empty((0))
        for index in np.atleast_1d(outlier_val):
            outlier_vector = np.zeros((out_params.shape[0], 1))
            outlier_vector[index] = 1
            out_params = np.hstack((out_params, outlier_vector))
        if detrend_poly:
            timepoints = out_params.shape[0]
            X = np.empty((timepoints, 0))
            for i in range(detrend_poly):
                X = np.hstack((X, legendre(
                    i + 1)(np.linspace(-1, 1, timepoints))[:, None]))
            out_params = np.hstack((out_params, X))
        filename = os.path.join(os.getcwd(), "filter_regressor%02d.txt" % idx)
        np.savetxt(filename, out_params, fmt=b"%.10f")
        out_files.append(filename)
    return out_files


def extract_noise_components(realigned_file, mask_file, num_components=5,
                             extra_regressors=None):
    """Derive components most reflective of physiological noise.
    Parameters
    ----------
    realigned_file: a 4D Nifti file containing realigned volumes
    mask_file: a 3D Nifti file containing white matter + ventricular masks
    num_components: number of components to use for noise decomposition
    extra_regressors: additional regressors to add
    Returns
    -------
    components_file: a text file containing the noise components
    """
    import os
    import nibabel as nb
    import numpy as np
    import scipy as sp
    from   nipype.utils.filemanip import filename_to_list

    imgseries = nb.load(realigned_file)
    components = None
    for filename in filename_to_list(mask_file):
        mask = nb.load(filename).get_data()
        if len(np.nonzero(mask > 0)[0]) == 0:
            continue
        voxel_timecourses = imgseries.get_data()[mask > 0]
        voxel_timecourses[np.isnan(np.sum(voxel_timecourses, axis=1)), :] = 0
        # remove mean and normalize by variance
        # voxel_timecourses.shape == [nvoxels, time]
        X = voxel_timecourses.T
        stdX = np.std(X, axis=0)
        stdX[stdX == 0] = 1.
        stdX[np.isnan(stdX)] = 1.
        stdX[np.isinf(stdX)] = 1.
        X = (X - np.mean(X, axis=0)) / stdX
        u, _, _ = sp.linalg.svd(X, full_matrices=False)
        if components is None:
            components = u[:, :num_components]
        else:
            components = np.hstack((components, u[:, :num_components]))
    if extra_regressors:
        regressors = np.genfromtxt(extra_regressors)
        components = np.hstack((components, regressors))
    components_file = os.path.join(os.getcwd(), 'noise_components.txt')

    np.savetxt(components_file, components, fmt=b"%.10f")
    return components_file



# -*- coding: utf-8 -*-
"""
Motion statistics calculation helper functions
This file is refactored from C-PAC:
https://github.com/FCP-INDI/C-PAC
"""

import nipype.pipeline.engine as pe
from   nipype.interfaces import Function, IdentityInterface

from   ..config import setup_node


def calc_friston_twenty_four(in_file):
    """ Method to calculate friston twenty four parameters.
    
    Parameters
    ----------
    in_file: string
        input movement parameters file from motion correction
    
    Returns
    -------
    new_file: string
        output 1D file containing 24 parameter values
    """
    import os.path as op
    import numpy as np

    data = np.genfromtxt(in_file)

    data_squared = data ** 2
    new_data = np.concatenate((data, data_squared), axis=1)

    data_roll = np.roll(data, 1, axis=0)
    data_roll[0] = 0
    new_data = np.concatenate((new_data, data_roll), axis=1)

    data_roll_squared = data_roll ** 2
    new_data = np.concatenate((new_data, data_roll_squared), axis=1)

    new_file = 'fristons_twenty_four.1D'
    np.savetxt(new_file, new_data, fmt='%0.8f', delimiter=' ')

    return op.abspath(new_file)


def fristons_twenty_four_wf(wf_name='fristons_twenty_four'):
    """ The main purpose of this workflow is to calculate 24 parameters including
    the 6 motion parameters of the current volume and the preceeding volume, 
    plus each of these values squared. 
    
    Parameters
    ----------
    wf_name: str
        Workflow name

    Returns 
    -------
    wf: workflow object

    Nipype Inputs
    -------------
    f24_input.in_file: str
        Path to the input movement file from motion correction.

    Nipype Outputs
    -------------
    f24_output.out_file: str
        Path to 1D file containing the friston 24 parameters.

    References
    ----------
 .. [1] Friston, K. J., Williams, S., Howard, R., Frackowiak, R. S., & Turner, R. (1996).
       Movement-related effects in fMRI time-series. Magnetic Resonance in Medicine, 35(3),346-355
    """
    wf = pe.Workflow(name=wf_name)

    # specify input and output fields
    in_fields  = ["in_file",]

    out_fields = ["out_file",]

    f24_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                           name='f24_input')

    calc_friston = setup_node(Function(input_names=['in_file'],
                                       output_names=['out_file'],
                                       function=calc_friston_twenty_four),
                             name='calc_friston')

    f24_output = setup_node(IdentityInterface(fields=out_fields),
                         name='f24_output')

    # Connect the nodes
    wf.connect([
                (f24_input,    calc_friston,  [("in_file",  "in_file")]),
                (calc_friston, f24_output,    [("out_file", "out_file")]),

              ])
    return wf


def motion_power_stats_wf(wf_name='gen_motion_stats'):
    """ The main purpose of this workflow is to get various statistical measures from the
    movement/motion parameters obtained in functional preprocessing.

    These parameters (FD calculations) are also required to carry out scrubbing.

    Order of commands:

    - Calculate Frame Wise Displacement FD as per power et al., 2012

      Differentiating head realignment parameters across frames yields a six dimensional timeseries that represents
      instantaneous head motion.
      Rotational displacements are converted from degrees to millimeters by calculating displacement on the surface of
      a sphere of radius 50 mm.[R5]

    - Calculate Frame wise Displacement FD as per jenkinson et al., 2002

    - Calculate Frames to exclude

      Remove all frames which are below the threshold

    - Calculate Frames to include

      Include all the frames which are above the threshold

    - Calculate DVARS

      DVARS (D temporal derivative of timecourses, VARS referring to RMS variance over voxels) indexes
      the rate of change of BOLD signal across the entire brain at each frame of data.To calculate
      DVARS, the volumetric timeseries is differentiated (by backwards differences) and RMS signal
      change is calculated over the whole brain.DVARS is thus a measure of how much the intensity
      of a brain image changes in comparison to the previous timepoint (as opposed to the global
      signal, which is the average value of a brain image at a timepoint).[R5]

    - Calculate Power parameters::

        MeanFD : Mean (across time/frames) of the absolute values for Framewise Displacement (FD),
        computed as described in Power et al., Neuroimage, 2012)

        rootMeanSquareFD : Root mean square (RMS; across time/frames) of the absolute values for FD

        NumFD >=threshold : Number of frames (time points) where movement (FD) exceeded threshold

        rmsFD : Root mean square (RMS; across time/frames) of the absolute values for FD

        FDquartile(top 1/4th FD) : Mean of the top 25% highest FD values

        PercentFD( > threshold) : Number of frames (time points) where movement (FD) exceeded threshold
                                  expressed as a percentage of the total number of frames (time points)

        MeanDVARS : Mean of voxel DVARS

    - Calculate Motion Parameters

      Following motion parameters are calculated::

        Subject, Scan, Mean Relative RMS Displacement, Max Relative RMS Displacement,
        Movements >threshold, Mean Relative Mean Rotation, Mean Relative Maxdisp,
        Max Relative Maxdisp, Max Abs Maxdisp, Max Relative Roll,Max Relative Pitch,
        Max Relative Yaw, Max Relative dS-I, Max Relative dL-R,Max Relative dP-A,
        Mean Relative Roll, Mean Relative Pitch,Mean Relative Yaw, Mean Relative dS-I,
        Mean Relative dL-R, Mean Relative dP-A, Max Abs Roll, Max Abs Pitch, Max Abs Yaw,
        Max Abs dS-I, Max Abs dL-R, Max Abs dP-A, Mean Abs Roll,Mean Abs Pitch,Mean Abs Yaw,
        Mean Abs dS-I,Mean Abs dL-R,Mean Abs dP-A

    Parameters
    ----------
    wf_name: workflow object
        Workflow name
    
    Returns 
    -------
    param_wf: workflow object
          Workflow object containing various movement/motion and power parameters estimates.  

    Nipype inputs
    -------------
    inputspec.motion_correct : string (func/rest file or a list of func/rest nifti file)
        Path to motion corrected functional data

    inputspec.mask : string (nifti file)
        Path to field contianing brain-only mask for the functional data

    inputspec.max_displacement : string (Mat file)
        maximum displacement (in mm) vector for brain voxels in each volume.
        This file is obtained in functional preprocessing step

    inputspec.movement_parameters : string (Mat file)
        1D file containing six movement/motion parameters(3 Translation, 3 Rotations)
        in different columns (roll pitch yaw dS  dL  dP), obtained in functional preprocessing step
        
    scrubbing_input.threshold : a float
        scrubbing threshold
        
    scrubbing_input.remove_frames_before : an integer
        count of preceding frames to the offending time
        frames to be removed (i.e.,those exceeding FD threshold)

    scrubbing_input.remove_frames_after : an integer
        count of subsequent frames to the offending time
        frames to be removed (i.e., those exceeding FD threshold)
            
    Nipype outputs
    --------------
    outputspec.FD_1D : 1D file
        mean Framewise Displacement (FD)

    outputspec.frames_ex_1D : 1D file
        Number of frames that would be censored ("scrubbed")
        also removing the offending time frames (i.e., those exceeding the threshold),
        the preceeding frame, and the two subsequent frames

    outputspec.frames_in_1D : 1d file
        Number of frames left after removing for scrubbing

    outputspec.power_params : txt file
        Text file various power parameters for scrubbing.

    outputspec.motion_params : txt file
       Text file containing various movement parameters

    References
    ----------
    .. [1] Power, J. D., Barnes, K. A., Snyder, A. Z., Schlaggar, B. L., & Petersen, S. E. (2012). Spurious
           but systematic correlations in functional connectivity MRI networks arise from subject motion. NeuroImage, 59(3),
           2142-2154. doi:10.1016/j.neuroimage.2011.10.018
           
    .. [2] Power, J. D., Barnes, K. A., Snyder, A. Z., Schlaggar, B. L., & Petersen, S. E. (2012). Steps
           toward optimizing motion artifact removal in functional connectivity MRI; a reply to Carp.
           NeuroImage. doi:10.1016/j.neuroimage.2012.03.017
    
    .. [3] Jenkinson, M., Bannister, P., Brady, M., Smith, S., 2002. Improved optimization for the robust 
           and accurate linear registration and motion correction of brain images. Neuroimage 17, 825-841.
    """
    wf = pe.Workflow(name=wf_name)

    # specify input and output fields
    in_fields  = ["in_files",
                  "anat",
                  "atlas_anat",
                  "coreg_target",
                  "tissues",
                  "lowpass_freq",
                  "highpass_freq",]

    out_fields = ["motion_corrected",
                  "motion_params",
                  "tissues",
                  "anat",
                  "time_filtered",
                  "smooth",
                  "time_filtered_mni",
                  "smooth_mni",
                  "tsnr_file",
                  "epi_brain_mask",
                  "tissues_brain_mask",
                  "motion_regressors",
                  "compcor_regressors",
                  "gsr_regressors",
                  "nuis_corrected",
                  "epi_mni",
                  "epi_mni_warpfield",]

    inputNode = setup_node(IdentityInterface(fields=['subject_id',
                                                       'scan_id',
                                                       'movement_parameters',
                                                       'max_displacement',
                                                       'motion_correct',
                                                       'mask',
                                                       'oned_matrix_save']),
                        name='inputspec')

    scrubbing_input = setup_node(IdentityInterface(fields=['threshold',
                                                              'remove_frames_before',
                                                              'remove_frames_after']),
                             name='scrubbing_input')

    outputNode = setup_node(IdentityInterface(fields=['FD_1D',
                                                        'FDJ_1D',
                                                        'frames_ex_1D',
                                                        'frames_in_1D',
                                                        'power_params',
                                                        'motion_params']),
                        name='outputspec')


    # calculate mean DVARS
    cal_DVARS = setup_node(Function(input_names=['in_file',
                                                   'mask'],
                                      output_names=['out_file'],
                                      function=calculate_DVARS),
                             name='cal_DVARS')
    pm.connect(inputNode, 'motion_correct', cal_DVARS, 'rest')
    pm.connect(inputNode, 'mask', cal_DVARS, 'mask')


    # Calculating mean Framewise Displacement as per power et al., 2012
    calculate_FD = setup_node(Function(input_names=['in_file'],
                                         output_names=['out_file'],
                                         function=calculate_FD_P),
                           name='calculate_FD')
    
    pm.connect(inputNode, 'movement_parameters', 
               calculate_FD, 'in_file' )
    
    pm.connect(calculate_FD, 'out_file', 
               outputNode, 'FD_1D')
    
    # Calculating mean Framewise Displacement as per jenkinson et al., 2002
    calculate_FDJ = setup_node(Function(input_names=['in_file'],
                                         output_names=['out_file'],
                                           function=calculate_FD_J),
                             name='calculate_FDJ')
    
    pm.connect(inputNode, 'oned_matrix_save', 
               calculate_FDJ, 'in_file' )
    
    pm.connect(calculate_FDJ, 'out_file', 
               outputNode, 'FDJ_1D')

    ##calculating frames to exclude and include after scrubbing
    exclude_frames = setup_node(Function(input_names=['in_file', 
                                                        'threshold',
                                                        'frames_before',
                                                        'frames_after'],
                                           output_names=['out_file'],
                                           function=set_frames_ex),
                             name='exclude_frames')

    pm.connect(calculate_FD, 'out_file', 
               exclude_frames, 'in_file')
    pm.connect(scrubbing_input, 'threshold', 
               exclude_frames, 'threshold')
    pm.connect(scrubbing_input, 'remove_frames_before',
               exclude_frames, 'frames_before')
    pm.connect(scrubbing_input, 'remove_frames_after',
               exclude_frames, 'frames_after')
    
    pm.connect(exclude_frames, 'out_file', 
               outputNode, 'frames_ex_1D')
    
    include_frames = setup_node(Function(input_names=['in_file',
                                                        'threshold', 
                                                        'exclude_list'],
                                           output_names=['out_file'],
                                           function=set_frames_in),
                             name='include_frames')
    pm.connect(calculate_FD, 'out_file', 
               include_frames, 'in_file')
    pm.connect(scrubbing_input, 'threshold', 
               include_frames, 'threshold')
    pm.connect(exclude_frames, 'out_file', 
               include_frames, 'exclude_list')

    pm.connect(include_frames, 'out_file', 
               outputNode, 'frames_in_1D')

    
    calc_motion_parameters = setup_node(Function(input_names=["subject_id", 
                                                                "scan_id", 
                                                                "movement_parameters",
                                                                "max_displacement"],
                                                   output_names=['out_file'],
                                                   function=gen_motion_parameters),
                                     name='calc_motion_parameters')
    pm.connect(inputNode, 'subject_id',
               calc_motion_parameters, 'subject_id')
    pm.connect(inputNode, 'scan_id',
               calc_motion_parameters, 'scan_id')
    pm.connect(inputNode, 'movement_parameters',
               calc_motion_parameters, 'movement_parameters')
    pm.connect(inputNode, 'max_displacement',
               calc_motion_parameters, 'max_displacement')
    
    pm.connect(calc_motion_parameters, 'out_file', 
               outputNode, 'motion_params')


    calc_power_parameters = setup_node(Function(input_names=["subject_id", 
                                                                "scan_id", 
                                                                "FD_1D",
                                                                "FDJ_1D", 
                                                                "threshold",
                                                                "DVARS"],
                                                   output_names=['out_file'],
                                                   function=gen_power_parameters),
                                     name='calc_power_parameters')
    pm.connect(inputNode, 'subject_id',
               calc_power_parameters, 'subject_id')
    pm.connect(inputNode, 'scan_id',
               calc_power_parameters, 'scan_id')
    pm.connect(cal_DVARS, 'out_file',
               calc_power_parameters, 'DVARS')
    pm.connect(calculate_FD, 'out_file',
               calc_power_parameters, 'FD_1D')
    pm.connect(calculate_FDJ, 'out_file',
               calc_power_parameters, 'FDJ_1D')
    pm.connect(scrubbing_input, 'threshold',
               calc_power_parameters, 'threshold')


    pm.connect(calc_power_parameters, 'out_file', 
               outputNode, 'power_params')

    return pm


def set_frames_ex(in_file, threshold, 
                  frames_before=1, frames_after=2):
    """
    Method to calculate Number of frames that would be censored 
    ("scrubbed") by removing the offending time frames 
    (i.e., those exceeding FD threshold), the preceding frame, 
    and the two subsequent frames
    
    Parameters
    ----------
    in_file : a string 
        framewise displacement(FD) file path
    threshold : a float
         scrubbing threshold value set in configuration file
    frames_before : an integer
        number of frames preceding the offending time frame
        by default value is 1
    frames_after : an integer
        number of frames following the offending time frame
        by default value is 2
         
    Returns
    -------
    out_file : string
        path to file containing offending time frames
    """

    import os
    import numpy as np
    from numpy import loadtxt

    out_file = os.path.join(os.getcwd(), 'frames_ex.1D')
    data = loadtxt(in_file)
    #masking zeroth timepoint value as 0, since the mean displacment value for
    #zeroth timepoint cannot be calculated, as there is no timepoint before it
    data[0] = 0

    extra_indices = []

    indices = [i[0] for i in (np.argwhere(data >= threshold)).tolist()]
    #print "initial indices-->", indices
    
    for i in indices:
        #remove preceding frames
        if i > 0 :
            count = 1
            while count <= frames_before:
                extra_indices.append(i-count)
                count+=1
                
        #remove following frames
        count = 1
        while count <= frames_after:
            extra_indices.append(i+count)
            count+=1
                    
    indices = list(set(indices) | set(extra_indices))
    indices.sort()

    f = open(out_file, 'a')

    #print "indices preceding and following -->", \
    #      set(extra_indices)
    #print "final indices -->", indices
    for idx in indices:
        f.write('%s,' % int(idx))

    f.close()

    return out_file


def calculate_FD_P(in_file):
    """ Method to calculate Framewise Displacement (FD) calculations (Power et al., 2012).
    Parameters
    ----------
    in_file: str
        Movement parameters vector file path.
    
    Returns
    -------
    out_file: str
        Frame-wise displacement matrix file path.
    """
    import os.path as op

    import numpy as np

    out_file = 'FD.1D'

    lines = open(in_file, 'r').readlines()
    rows = [[float(x) for x in line.split()] for line in lines]
    cols = np.array([list(col) for col in zip(*rows)])
    
    translations = np.transpose(np.abs(np.diff(cols[0:3, :])))
    rotations = np.transpose(np.abs(np.diff(cols[3:6, :])))

    FD_power = np.sum(translations, axis = 1) + (50*3.141/180)*np.sum(rotations, axis =1)
    
    #FD is zero for the first time point
    FD_power = np.insert(FD_power, 0, 0)
    
    np.savetxt(out_file, FD_power)
    
    return op.abspath(out_file)
    

def calculate_FD_J(in_file):
    """
    Method to calculate Framewise Displacement (FD) calculations (Jenkinson et al., 2002).
    Parameters
    ----------
    in_file: str
        movement parameters vector file path.

    Returns
    -------
    out_file: str
        Frame-wise displacement matrix file path.

    Notes
    -----
    `in_file` should have one 3dvolreg affine matrix in one row - NOT the motion parameters.
    """
    import os.path as op
    import sys
    import math

    import numpy as np

    out_file = 'FD_J.1D'

    #print in_file
    pm_ = np.genfromtxt(in_file)

    pm = np.zeros((pm_.shape[0],pm_.shape[1]+4))
    pm[:,:12]=pm_
    pm[:,12:]=[0.0, 0.0, 0.0, 1.0]

    flag = 0

    #The default radius (as in FSL) of a sphere represents the brain
    rmax = 80.0

    with open(out_file, 'w') as f:
        #rigid body transformation matrix
        T_rb_prev = np.matrix(np.eye(4))

        for i in range(0, pm.shape[0]):
            T_rb = np.matrix(pm[i].reshape(4,4)) # making use of the fact that the order of aff12 matrix is "row-by-row"

            if flag == 0:
                flag = 1
                # first timepoint
                print >> f, 0
            else:
                M = np.dot(T_rb, T_rb_prev.I) - np.eye(4)
                A = M[0:3, 0:3]
                b = M[0:3, 3]

                FD_J = math.sqrt((rmax*rmax/5)*np.trace(np.dot(A.T, A)) + np.dot(b.T, b))
                print >> f, '%.8f'%FD_J

            T_rb_prev = T_rb

    return op.abspath(out_file)


def set_frames_in(in_file, threshold, exclude_list):
    """ Method to calculate the frames that are left after censoring for scrubbing.
     
     Parameters
     ----------
     in_file : string
        framewise displacement(FD) file path
     threshold : float
        scrubbing thereshold set in configuration file
     exclude_list : string
        path of file containing sensored timepoints
    
     Returns
     -------
     out_file : string 
        path of file containing remaining uncensored timepoints 
    """
    import os.path as op

    import numpy as np

    out_file = 'frames_in.1D'
    data = np.loadtxt(in_file)

    #masking zeroth timepoint value as 0, since the mean displacment value for
    #zeroth timepoint cannot be calculated, as there is no timepoint before it
    data[0] = 0
    indices = [i[0] for i in (np.argwhere(data < threshold)).tolist()]

    indx = []
    with open(exclude_list, 'r') as f:
        line = f.readline()

    if line:
        line = line.strip(',')
        indx = map(int, line.split(","))

    if indx:
        indices = list(set(indices) - set(indx))

    with open(out_file, 'a') as f:
        for idx in indices:
            f.write('%s,' % int(idx))

    return op.abspath(out_file)


def gen_motion_parameters(subject_id, scan_id, movement_parameters, max_displacement):
    """  Method to calculate all the movement parameters.
    
    Parameters
    ----------
    subject_id : string
        subject name or id

    scan_id : string
        scan name or id

    max_displacement : string
        path of file with maximum displacement (in mm) for brain voxels in each volume    

    movement_parameters : string
        path of 1D file containing six movement/motion parameters(3 Translation, 
        3 Rotations) in different columns (roll pitch yaw dS  dL  dP)
    
    Returns 
    -------
    out_file : string 
        path to csv file contianing various motion parameters
    """

    import os.path as op
    import numpy as np
    import re

    out_file = 'motion_parameters.txt'

    with open(out_file, 'w') as f:
        #f.write(str(os.getcwd()))   ### work
        header = "Subject,Scan,Mean_Relative_RMS_Displacement," \
            "Max_Relative_RMS_Displacement,Movements_gt_threshold,"\
            "Mean_Relative_Mean_Rotation,Mean_Relative_Maxdisp,Max_Relative_Maxdisp," \
            "Max_Abs_Maxdisp,Max Relative_Roll,Max_Relative_Pitch," \
            "Max_Relative_Yaw,Max_Relative_dS-I,Max_Relative_dL-R," \
            "Max_Relative_dP-A,Mean_Relative_Roll,Mean_Relative_Pitch,Mean_Relative_Yaw," \
            "Mean_Relative_dS-I,Mean_Relative_dL-R,Mean_Relative_dP-A,Max_Abs_Roll," \
            "Max_Abs_Pitch,Max_Abs_Yaw,Max_Abs_dS-I,Max_Abs_dL-R,Max_Abs_dP-A," \
            "Mean_Abs_Roll,Mean_Abs_Pitch,Mean_Abs_Yaw,Mean_Abs_dS-I,Mean_Abs_dL-R,Mean_Abs_dP-A\n"

        f.write(header)

        f.write("%s," % (subject_id))
        f.write("%s," % (scan_id))

        arr = np.genfromtxt(movement_parameters)
        arr = arr.T

        ##Relative RMS of translation
        rms = np.sqrt(arr[3]*arr[3] + arr[4]*arr[4] + arr[5]*arr[5])
        diff = np.diff(rms)
        MEANrms = np.mean(abs(diff))
        f.write("%.3f," % (MEANrms))

        #Max Relative RMS Displacement
        MAXrms = np.max(abs(diff))
        f.write("%.3f," % (MAXrms))

        ##NUMBER OF relative RMS movements >0.1mm
        NUMmove = np.sum(abs(diff) > 0.1)
        f.write("%.3f," % (NUMmove))

        ##Mean of mean relative rotation (params 1-3)
        MEANrot = np.mean(np.abs(np.diff((abs(arr[0])+ abs(arr[1])+ abs(arr[2]))/3 ) ) )
        f.write("%.3f," % (MEANrot))

        with open(max_displacement, 'r') as disp_f:
            lines = disp_f.readlines()

        list1 = []
        #remove any other information aother than matrix from
        #max displacement file. afni adds infomration to the file
        for l in lines:
            if re.match("^\d+?\.\d+?$", l.strip()):
                list1.append(float(l.strip()))

        arr2 = np.array(list1, dtype='float')

        #Mean Relative Maxdisp
        mean = np.mean(np.diff(arr2))
        f.write("%.3f," % (mean))

        #Max Relative Maxdisp
        relMAX = np.max(abs(np.diff(arr2)))
        f.write("%.3f," % (relMAX))

        #Max Abs Maxdisp
        MAX= np.max(arr2)
        f.write("%.3f," %(MAX))

        #Max Relative Roll,Max Relative Pitch,
        #Max Relative Yaw,Max Relative dS-I,
        #Max Relative dL-R,Max Relative dP-A
        for i in range(6):
            f.write("%.6f," %(np.max(abs(np.diff(arr[i])))))

        #Mean Relative Roll,Mean Relative Pitch,
        #Mean Relative Yaw,Mean Relative dS-I,
        #Mean Relative dL-R,Mean Relative dP-A
        for i in range(6):
            f.write("%.6f," %(np.mean(np.diff(arr[i]))))

        #Max Abs Roll,Max Abs Pitch,Max Abs Yaw,
        #Max Abs dS-I,Max Abs dL-R,Max Abs dP-A
        for i in range(6):
            f.write("%.6f," %(np.max(abs(arr[i]))))

        #Mean Abs Roll,Mean Abs Pitch,Mean Abs Yaw,
        #Mean Abs dS-I,Mean Abs dL-R,Mean Abs dP-A
        for i in range(6):
            f.write("%.6f," %(np.mean(abs(arr[i]))))

    return op.abspath(out_file)


def gen_power_parameters(subject_id, scan_id, FD_1D, FDJ_1D, DVARS, threshold = 1.0):
    """ Method to generate Power parameters for scrubbing
    
    Parameters
    ----------
    subject_id : string
        subject name or id

    scan_id : string
        scan name or id

    FD_ID: string
        framewise displacement(FD as per power et al., 2012) file path

    FDJ_ID: string
        framewise displacement(FD as per jenkinson et al., 2002) file path

    threshold : float
        scrubbing threshold set in the configuration
        by default the value is set to 1.0

    DVARS : string
        path to numpy file containing DVARS
    
    Returns
    -------
    out_file : string (csv file)
        path to csv file containing all the pow parameters 
    """
    import os.path as op
    import numpy as np

    out_file = 'pow_params.txt'

    with open(out_file,'w') as f:

        header = "Subject,Scan,MeanFD,MeanFD_Jenkinson," \
        "NumFD_greater_than_%.2f,rootMeanSquareFD,FDquartile(top1/4thFD)," \
        "PercentFD_greater_than_%.2f,MeanDVARS\n".format(threshold,threshold)

        f.write(header)

        f.write("%s," % subject_id)
        f.write("%s," % scan_id)

        powersFD_data = np.loadtxt(FD_1D)
        jenkFD_data   = np.loadtxt(FDJ_1D)

        #Mean (across time/frames) of the absolute values
        #for Framewise Displacement (FD)
        meanFD  = np.mean(powersFD_data)
        f.write('%.4f,' % meanFD)

        #Mean FD Jenkinson
        meanFD_Jenkinson = np.mean(jenkFD_data)
        f.write('%.4f,' % meanFD_Jenkinson)

        #Number of frames (time points) where movement
        #(FD) exceeded threshold
        numFD = float(jenkFD_data[jenkFD_data > threshold].size)
        f.write('%.4f,' % numFD)

        #Root mean square (RMS; across time/frames)
        #of the absolute values for FD
        rmsFD = np.sqrt(np.mean(jenkFD_data))
        f.write('%.4f,' % rmsFD)

        #Mean of the top quartile of FD is $FDquartile
        quat=int(len(jenkFD_data)/4)
        FDquartile=np.mean(np.sort(jenkFD_data)[::-1][:quat])
        f.write('%.4f,' % FDquartile)

        ##NUMBER OF FRAMES >threshold FD as percentage of total num frames
        count = np.float(jenkFD_data[jenkFD_data>threshold].size)
        percentFD = (count*100/(len(jenkFD_data)+1))
        f.write('%.4f,' %percentFD)

        #Mean DVARS
        meanDVARS = np.mean(np.load(DVARS))
        f.write('%.4f,' % meanDVARS)

    return op.abspath(out_file)


def calculate_DVARS(rest, mask):
    """
    Method to calculate DVARS as per power's method
    
    Parameters
    ----------
    rest : string (nifti file)
        path to motion correct functional data

    mask : string (nifti file)
        path to brain only mask for functional data
        
    Returns
    -------
    out_file : string (numpy mat file)
        path to file containing  array of DVARS 
        calculation for each voxel
    """
    import os.path as op

    import numpy as np
    import nibabel as nib

    out_file = 'DVARS.npy'
    
    rest_data = nib.load(rest).get_data().astype(np.float32)
    mask_data = nib.load(mask).get_data().astype('bool')
    
    #square of relative intensity value for each voxel across
    #every timepoint 
    data = np.square(np.diff(rest_data, axis = 3))

    #applying mask, getting the data in the brain only
    data = data[mask_data]

    #square root and mean across all timepoints inside mask
    DVARS = np.sqrt(np.mean(data, axis=0))

    np.save(out_file, DVARS)
    
    return op.abspath(out_file)

    



# -*- coding: utf-8 -*-
"""
Nipype registration nodes and workflows
"""
import os.path as path

import nipype.pipeline.engine as pe
from   nipype.algorithms.misc import Gunzip
from   nipype.interfaces.base import traits, isdefined
from   nipype.interfaces.utility import IdentityInterface, Function
import nipype.interfaces.afni as afni
import nipype.interfaces.fsl as fsl
import nipype.interfaces.spm  as spm

from ..interfaces.nilearn import mean_img, concat_imgs
from .spatial import get_bounding_box
from ..config import setup_node, get_config_setting
from ..utils import spm_tpm_priors_path


def spm_apply_deformations(in_file=traits.Undefined, trans_field=traits.Undefined, bbox=traits.Undefined,
                           voxel_sizes=None):
    """Return a Normalize12 interface object.
    SPM12's new Normalise routine for warping an image to a template.
    For more info:
    http://www.mit.edu/~satra/nipype-nightly/interfaces/generated/nipype.interfaces.spm.preprocess.html#normalize12

    Parameters
    ----------
    in_file: file

    trans_field: file
        file y_*.nii containing 3 deformation fields for the deformation in
        x, y and z dimension

    bbox: (a list of from 2 to 2 items which are a list of
           from 3 to 3 items which are a float)
        write_bounding_box option.
        3x2-element list of lists representing the bounding box (in mm) to
        be written
        See the preproc.get_bounding_box function.
        This is important to set when applying deformation fields
        to cropped images.

    voxel_sizes: list of 3 ints or floats
        Default: [1, 1, 1]

    Nipype Ouputs
    -------------
    deformation_field: (a list of items which are an existing file name)
        NIfTI file containing 3 deformation fields for the deformation in x,
        y and z dimension

    normalized_files: (a list of items which are an existing file name)
        Normalized other files

    normalized_image: (a list of items which are an existing file name)
        Normalized file that needed to be aligned

    """
    norm12 = spm.Normalize12(jobtype='write')

    if voxel_sizes is not None:
        norm12.inputs.write_voxel_sizes = voxel_sizes

    norm12.inputs.deformation_file   = trans_field
    norm12.inputs.image_to_align     = in_file
    norm12.inputs.write_bounding_box = bbox

    #norm12.run()
    return norm12


def spm_normalize(in_file=traits.Undefined, template=None, **kwargs):
    """Return a Normalize12 interface object.
    SPM12's new Normalise routine for warping an image to a template.

    Parameters
    ----------
    in_file: file

    template: file
        Template in form of tissue probablitiy maps to normalize to
        mutually_exclusive: deformation_file.
        Default: the SPM TPM file.

    Nipype Ouputs
    -------------
    deformation_field: (a list of items which are an existing file name)
        NIfTI file containing 3 deformation fields for the deformation in x,
        y and z dimension

    normalized_files: (a list of items which are an existing file name)
        Normalized other files

    normalized_image: (a list of items which are an existing file name)
        Normalized file that needed to be aligned

    Raises
    ------
    FileNotFoundError
        If `template` is `None` and can't find the TPM.nii file from SPM.

    Notes
    -----
    For more info:
    http://www.mit.edu/~satra/nipype-nightly/interfaces/generated/nipype.interfaces.spm.preprocess.html#normalize12
    """
    if template is None:
        template = spm_tpm_priors_path()

    norm12 = spm.Normalize12(jobtype='estwrite',
                             tpm=template,
                             **kwargs)

    if isdefined(in_file):
        norm12.inputs.image_to_align = in_file

    #norm12.run()
    return norm12


def spm_coregister(src_img=traits.Undefined, tgt_img=traits.Undefined, cost_function='mi'):
    """Use spm_coreg for estimating cross-modality rigid body alignment.
    The write_interp option is set to 0 by default.

    More info: http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=39
    Parameters
    ----------
    src_img: str
        Path to the coregistration source image

    tgt_img: str
        Path to the coregistration target image

    cost_function: ('mi' or 'nmi' or 'ecc' or 'ncc')
        cost function, one of: 'mi' - Mutual Information,
         'nmi' - Normalised Mutual Information,
         'ecc' - Entropy Correlation Coefficient,
         'ncc' - Normalised Cross Correlation

        For more info:
        http://www.mit.edu/~satra/nipype-nightly/interfaces/generated/nipype.interfaces.spm.preprocess.html#coregister

    Returns
    -------
    coreg: nipype.interfaces.smp.Coregister
        spm.Coregister interface object
    """
    coreg = spm.Coregister()

    coreg.inputs.source = src_img
    coreg.inputs.target = tgt_img

    coreg.inputs.cost_function = cost_function
    coreg.inputs.jobtype = 'estwrite'

    #coreg.run()
    return coreg


def afni_deoblique(in_file=traits.Undefined, out_file=traits.Undefined, out_type='NIFTI_GZ'):
    """ Return a nipype interface for AFNI '3dWarp -deoblique'.

    Parameters
    ----------
    in_file: str
        Path to the input file

    out_file: str
        Path to the output file.

    out_type: str
        ('NIFTI_GZ' or 'AFNI' or 'NIFTI')
        AFNI output filetype

    Returns
    -------
    deob: nipype.interfaces.afni.Warp
    """
    deob = afni.Warp()
    deob.inputs.in_file = in_file
    deob.inputs.deoblique = True
    deob.inputs.out_file = out_file
    deob.inputs.outputtype = out_type

    return deob


def spm_warp_to_mni(wf_name="spm_warp_to_mni"):
    """ Run Gunzip and SPM Normalize12 to the list of files input and outputs the list of warped files.

    It does:
    - Warp each individual input image to the standard SPM template

    Parameters
    ----------
    wf_name: str
        Name of the workflow.

    Nipype Inputs
    -------------
    warp_input.in_files: list of traits.File
        The raw NIFTI_GZ image files

    Nipype outputs
    --------------
    warp_output.warped_files: list of existing file
        The warped files.

    Returns
    -------
    wf: nipype Workflow
    """
    # input
    # check if spm_pet_preproc.do_petpvc is True
    in_fields  = ["in_files"]
    out_fields = ["warped_files"]

    input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                       name="warp_input",)

    gunzip = pe.MapNode(Gunzip(), name="gunzip", iterfield=['in_file'])

    warp = setup_node(spm.Normalize12(jobtype='estwrite', affine_regularization_type='mni'),
                      name="normalize12", type="map", iterfield=['image_to_align'])

    # output
    output = setup_node(IdentityInterface(fields=out_fields),
                        name="warp_output")

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    wf.connect([
                # inputs
                (input,   gunzip, [("in_files", "in_file")]),
                (gunzip,  warp,   [("out_file", "image_to_align")]),

                # output
                (warp,    output, [("normalized_image", "warped_files")]),
               ])

    return wf


def spm_create_group_template_wf(wf_name="spm_create_group_template"):
    """ Pick all subject files in `grptemplate_input.in_files`, calculate an average
    image and smooth it with `"{}_smooth".format(wf_name)` node (you can configure the smooth `fwhm` from
    a config file.).

    It does:
    - calculate a mean image (across subjects) and
    - smooth it with a 8x8x8mm^3 gaussian kernel -> the result of this is the template.
    The size of the isometric smoothing gaussian kernel is given by one integer for the
    "{}_smooth.fwhm".format(wf_name) setting.

    You can also avoid calculating the mean image across subjects and setting a specific group template file by
    setting the configuration "{}.template_file".format(wf_name) to the path of the file you want.
    This image will be smoothed and used as a common template for the further pipeline steps.

    Parameters
    ----------
    wf_name: str
        Name of the workflow.

    Nipype Inputs
    -------------
    grptemplate_input.in_files: list of traits.File
        The raw NIFTI_GZ PET image files

    Nipype outputs
    --------------
    grptemplate_output.template: existing file
        The common custom PET template file.

    Returns
    -------
    wf: nipype Workflow
    """
    # input
    input = setup_node(IdentityInterface(fields=["in_files"]), name="grptemplate_input",)

    # checking if a template file has been set already
    template_file = get_config_setting("{}.template_file".format(wf_name))

    use_common_template = path.exists(template_file)
    if not use_common_template:
        # merge
        concat = setup_node(Function(function=concat_imgs,
                                     input_names=["in_files"],
                                     output_names=["out_file"],
                                     imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                            name='merge_time')

        # average
        average = setup_node(Function(function=mean_img,
                                      input_names=["in_file", "out_file"],
                                      output_names=["out_file"],
                                      imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                            name='group_average')
        average.inputs.out_file = 'group_average.nii.gz'

    #TODO: check what is the difference between nilearn.image.smooth_img and FSL IsotropicSmooth
    # smooth
    #smooth = setup_node(Function(function=smooth_img,
    #                             input_names=["in_file", "fwhm"],
    #                             output_names=["out_file"],
    #                             imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
    #                     name="{}_smooth".format(wf_name))
    smooth = setup_node(fsl.IsotropicSmooth(fwhm=8), name="{}_smooth".format(wf_name))

    # output
    output = setup_node(IdentityInterface(fields=["template"]), name="grptemplate_output",)

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # if I have to create the group template
    if not use_common_template:
        wf.connect([
                    # input
                    (input,       concat,  [("in_files", "in_files")]),

                    # merge, average and smooth
                    (concat,      average, [("out_file", "in_file")]),
                    (average,     smooth,  [("out_file", "in_file")]),

                    # output
                    (smooth,     output,   [("out_file", "template")]),
                   ])
    else: # if the template has been specified in the configuration file
        wf.add_nodes([input])

        smooth.inputs.in_file = template_file

        wf.connect([
                    # output
                    (smooth,     output,   [("out_file", "template")]),
                   ])

    return wf


def spm_register_to_template_wf(wf_name="spm_registration_to_template"):
    """Return a workflow that registers each reg_input.in_file to the file in reg_input.template.
    For now this does not do atlas registration.

    It does:
    - SPM12 Warp input image to the given template

    Parameters
    ----------
    wf_name: str
        Name of the workflow.

    Nipype Inputs
    -------------
    reg_input.in_file: traits.File
        The raw NIFTI_GZ subject image file.

    reg_input.template: list of traits.File
        The template file for inter-subject registration reference.

    Nipype outputs
    --------------
    reg_output.warped: existing file
        Image normalized to the given template.

    reg_output.warp_field: existing files
        Spatial normalization parameters .mat file.

    Returns
    -------
    wf: nipype Workflow
    """
    # specify input and output fields
    in_fields  = ["in_file",
                  "template",]

    out_fields = ["warped",
                  "warp_field",]

    # input
    reg_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                           name="reg_input")

    # warp each subject to the group template
    gunzip_template = setup_node(Gunzip(), name="gunzip_template",)
    gunzip_input    = setup_node(Gunzip(), name="gunzip_input",)

    warp2template = setup_node(spm.Normalize(jobtype="estwrite", out_prefix="wgrptemplate_"),
                               name="warp2template")

    get_bbox = setup_node(Function(function=get_bounding_box,
                                   input_names=["in_file"],
                                   output_names=["bbox"]),
                          name="get_bbox")

    # output
    reg_output = setup_node(IdentityInterface(fields=out_fields), name="reg_output")

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    wf.connect([
                # get template bounding box to apply to results
                (reg_input,     get_bbox,        [("template", "in_file")]),

                # gunzip some inputs
                (reg_input,     gunzip_input,    [("in_file",  "in_file")]),
                (reg_input,     gunzip_template, [("template", "in_file")]),

                # prepare the target parameters of the warp to template
                (gunzip_template, warp2template, [("out_file", "template")]),
                (get_bbox,        warp2template, [("bbox",     "write_bounding_box")]),

                # directly warp pet to the template
                (gunzip_input,    warp2template, [("out_file", "source")]),

                # output
                (warp2template, reg_output, [("normalization_parameters", "warp_field"),
                                             ("normalized_source",        "warped"),
                                             ]),
               ])

    return wf



# -*- coding: utf-8 -*-
"""
PET-MR image preprocessing nipype workflows.
"""
import os.path as op

import nipype.pipeline.engine    as pe
from   nipype.algorithms.misc    import Gunzip
from   nipype.interfaces.utility import Merge, IdentityInterface, Function
from   nipype.interfaces         import spm

from   .pvc      import petpvc_workflow
from   ..config  import (setup_node,
                        check_atlas_file,
                        get_config_setting)

from   ..preproc import (spm_normalize,
                         spm_coregister,
                         spm_apply_deformations,
                         get_bounding_box)

from   ..utils import (get_datasink,
                       spm_tpm_priors_path,
                       extend_trait_list,
                       get_input_node,
                       get_interface_node,
                       remove_ext,
                       get_input_file_name,
                       extension_duplicates)

from   .._utils import format_pair_list, flatten_list


# TODO: merge the two workflows below, maybe splitting them in
# two wf steps: pre-processing then registration.
def spm_mrpet_preprocessing(wf_name="spm_mrpet_preproc"):
    """ Run the PET pre-processing workflow against the
    gunzip_pet.in_file files.
    It depends on the anat_preproc_workflow, so if this
    has not been run, this function will run it too.

    # TODO: organize the anat2pet hack/condition somehow:
    If anat2pet:
    - SPM12 Coregister T1 and tissues to PET
    - PVC the PET image in PET space
    - SPM12 Warp PET to MNI
    else:
    - SPM12 Coregister PET to T1
    - PVC the PET image in anatomical space
    - SPM12 Warp PET in anatomical space to MNI through the
    `anat_to_mni_warp`.

    Parameters
    ----------
    wf_name: str
        Name of the workflow.

    Nipype Inputs
    -------------
    pet_input.in_file: traits.File
        The raw NIFTI_GZ PET image file

    pet_input.anat: traits.File
        Path to the high-contrast anatomical image.
        Reference file of the warp_field, i.e., the
        anatomical image in its native space.

    pet_input.anat_to_mni_warp: traits.File
        The warp field from the transformation of the
        anatomical image to the standard MNI space.

    pet_input.atlas_anat: traits.File
        The atlas file in anatomical space.

    pet_input.tissues: list of traits.File
        List of tissues files from the New Segment process.
        At least the first 3 tissues must be present.

    Nipype outputs
    --------------
    pet_output.pvc_out: existing file
        The results of the PVC process

    pet_output.brain_mask: existing file
        A brain mask calculated with the tissues file.

    pet_output.coreg_ref: existing file
        The coregistered reference image to PET space.

    pet_output.coreg_others: list of existing files
        List of coregistered files from coreg_pet.apply_to_files

    pet_output.pvc_warped: existing file
        Results from PETPVC normalized to MNI.
        The result of every internal pre-processing step
        is normalized to MNI here.

    pet_output.warp_field: existing files
        Spatial normalization parameters .mat files

    pet_output.gm_norm: existing file
        The output of the grey matter intensity
        normalization process.
        This is the last step in the PET signal correction,
        before registration.

    pet_output.atlas_pet: existing file
        Atlas image warped to PET space.
        If the `atlas_file` option is an existing file and
        `normalize_atlas` is True.

    Returns
    -------
    wf: nipype Workflow
    """
    # specify input and output fields
    in_fields  = ["in_file",
                  "anat",
                  "anat_to_mni_warp",
                  "tissues",]

    out_fields = ["brain_mask",
                  "coreg_others",
                  "coreg_ref",
                  "pvc_warped",
                  "pet_warped", # 'pet_warped' is a dummy entry to keep the fields pattern.
                  "warp_field",
                  "pvc_out",
                  "pvc_mask",
                  "gm_norm",]

    do_atlas, _ = check_atlas_file()
    if do_atlas:
        in_fields  += ["atlas_anat"]
        out_fields += ["atlas_pet" ]

    # input
    pet_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                           name="pet_input")

    # workflow to perform partial volume correction
    petpvc    = petpvc_workflow(wf_name="petpvc")

    merge_list = setup_node(Merge(4), name='merge_for_unzip')
    gunzipper = pe.MapNode(Gunzip(), name="gunzip", iterfield=['in_file'])

    warp_pet = setup_node(spm_normalize(), name="warp_pet")

    tpm_bbox = setup_node(Function(function=get_bounding_box,
                                   input_names=["in_file"],
                                   output_names=["bbox"]),
                          name="tpm_bbox")
    tpm_bbox.inputs.in_file = spm_tpm_priors_path()

    # output
    pet_output = setup_node(IdentityInterface(fields=out_fields), name="pet_output")

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # check how to perform the registration, to decide how to build the pipeline
    anat2pet = get_config_setting('registration.anat2pet', False)
    if anat2pet:
        wf.connect([
                    # inputs
                    (pet_input, petpvc,     [("in_file", "pvc_input.in_file"),
                                             ("anat",    "pvc_input.reference_file"),
                                             ("tissues", "pvc_input.tissues")]),

                    # gunzip some files for SPM Normalize
                    (petpvc,    merge_list, [("pvc_output.pvc_out",    "in1"),
                                             ("pvc_output.brain_mask", "in2"),
                                             ("pvc_output.gm_norm",    "in3")]),
                    (pet_input, merge_list, [("in_file",               "in4")]),

                    (merge_list, gunzipper, [("out", "in_file")]),

                    # warp the PET PVCed to MNI
                    (petpvc,    warp_pet,   [("pvc_output.coreg_ref", "image_to_align")]),
                    (gunzipper, warp_pet,   [("out_file",             "apply_to_files")]),
                    (tpm_bbox,  warp_pet,   [("bbox",                 "write_bounding_box")]),

                    # output
                    (petpvc,    pet_output, [("pvc_output.pvc_out",      "pvc_out"),
                                             ("pvc_output.brain_mask",   "brain_mask"),
                                             ("pvc_output.coreg_ref",    "coreg_ref"),
                                             ("pvc_output.coreg_others", "coreg_others"),
                                             ("pvc_output.gm_norm",      "gm_norm")]),

                    # output
                    (warp_pet,  pet_output, [("normalized_files",  "pvc_warped"),
                                             ("deformation_field", "warp_field")]),
                   ])
    else: # PET 2 ANAT
        collector  = setup_node(Merge(2), name='merge_for_warp')
        apply_warp = setup_node(spm_apply_deformations(), name="warp_pet")

        wf.connect([
                    # inputs
                    (pet_input, petpvc,     [("in_file", "pvc_input.in_file"),
                                             ("anat",    "pvc_input.reference_file"),
                                             ("tissues", "pvc_input.tissues")]),

                    # gunzip some files for SPM Normalize
                    (petpvc,    merge_list, [("pvc_output.pvc_out",    "in1"),
                                             ("pvc_output.brain_mask", "in2"),
                                             ("pvc_output.gm_norm",    "in3")]),
                    (pet_input, merge_list, [("in_file",               "in4")]),

                    (merge_list, gunzipper, [("out",                   "in_file")]),

                    # warp the PET PVCed to MNI
                    (gunzipper,   collector,   [("out_file",             "in1")]),
                    (petpvc,      collector,   [("pvc_output.coreg_ref", "in2")]),

                    (pet_input,   apply_warp,  [("anat_to_mni_warp", "deformation_file")]),
                    (collector,   apply_warp,  [("out",              "apply_to_files")]),
                    (tpm_bbox,    apply_warp,  [("bbox",             "write_bounding_box")]),

                    # output
                    (petpvc,    pet_output, [("pvc_output.pvc_out",      "pvc_out"),
                                             ("pvc_output.brain_mask",   "brain_mask"),
                                             ("pvc_output.petpvc_mask",  "petpvc_mask"),
                                             ("pvc_output.coreg_ref",    "coreg_ref"),
                                             ("pvc_output.coreg_others", "coreg_others"),
                                             ("pvc_output.gm_norm",      "gm_norm")]),

                    # output
                    (apply_warp,  pet_output, [("normalized_files",  "pvc_warped"),
                                               ("deformation_field", "warp_field")]),
                   ])


    if do_atlas:
        coreg_atlas = setup_node(spm_coregister(cost_function="mi"), name="coreg_atlas")

        # set the registration interpolation to nearest neighbour.
        coreg_atlas.inputs.write_interp = 0
        wf.connect([
                    (pet_input,   coreg_atlas, [("anat",                 "source")]),
                    (petpvc,      coreg_atlas, [("pvc_output.coreg_ref", "target")]),
                    (pet_input,   coreg_atlas, [("atlas_anat",           "apply_to_files")]),
                    (coreg_atlas, pet_output,  [("coregistered_files",   "atlas_pet")]),
        ])

    return wf


def spm_mrpet_grouptemplate_preprocessing(wf_name="spm_mrpet_grouptemplate_preproc"):
    """ Run the PET pre-processing workflow against the gunzip_pet.in_file files.
    It depends on the anat_preproc_workflow, so if this has not been run, this function
    will run it too.

    This is identical to the workflow defined in `spm_mrpet_preprocessing`,
    with the only difference that we now normalize all subjects agains a custom
    template using the spm Old Normalize interface.

    It does:
    - SPM12 Coregister T1 and tissues to PET
    - PVC the PET image in PET space
    - SPM12 Warp PET to the given template

    Parameters
    ----------
    wf_name: str
        Name of the workflow.

    Nipype Inputs
    -------------
    pet_input.in_file: traits.File
        The raw NIFTI_GZ PET image file.

    pet_input.atlas_anat: traits.File
        The atlas file in anatomical space.

    pet_input.anat: traits.File
        Path to the high-contrast anatomical image.
        Reference file of the warp_field, i.e., the anatomical image in its native space.

    pet_input.tissues: list of traits.File
        List of tissues files from the New Segment process. At least the first
        3 tissues must be present.

    pet_input.pet_template: traits.File
        The template file for inter-subject registration reference.

    Nipype outputs
    --------------
    pet_output.pvc_out: existing file
        The results of the PVC process.

    pet_output.brain_mask: existing file
        A brain mask calculated with the tissues file.

    pet_output.coreg_ref: existing file
        The coregistered reference image to PET space.

    pet_output.coreg_others: list of existing files
        List of coregistered files from coreg_pet.apply_to_files.

    pet_output.pet_warped: existing file
        PET image normalized to the group template.

    pet_output.pvc_warped: existing file
        The outputs of the PETPVC workflow normalized to the group template.
        The result of every internal pre-processing step is normalized to the
        group template here.

    pet_output.warp_field: existing files
        Spatial normalization parameters .mat files.

    pet_output.gm_norm: existing file
        The output of the grey matter intensity normalization process.
        This is the last step in the PET signal correction, before registration.

    pet_output.atlas_pet: existing file
        Atlas image warped to PET space.
        If the `atlas_file` option is an existing file and `normalize_atlas` is True.

    Returns
    -------
    wf: nipype Workflow
    """
    # specify input and output fields
    in_fields  = ["in_file",
                  "anat",
                  "tissues",
                  "pet_template"]

    out_fields = ["brain_mask",
                  "coreg_others",
                  "coreg_ref",
                  "pvc_warped",
                  "pet_warped",
                  "warp_field",
                  "pvc_out",
                  "pvc_mask",
                  "gm_norm",]

    do_atlas, _ = check_atlas_file()
    if do_atlas:
        in_fields  += ["atlas_anat"]
        out_fields += ["atlas_pet" ]

    # input
    pet_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                           name="pet_input")

    # workflow to perform partial volume correction
    petpvc = petpvc_workflow(wf_name="petpvc")

    unzip_mrg = setup_node(Merge(4), name='merge_for_unzip')
    gunzipper = pe.MapNode(Gunzip(), name="gunzip", iterfield=['in_file'])

    # warp each subject to the group template
    gunzip_template = setup_node(Gunzip(), name="gunzip_template",)
    gunzip_pet      = setup_node(Gunzip(), name="gunzip_pet",)

    warp_mrg = setup_node(Merge(2), name='merge_for_warp')
    warp2template = setup_node(spm.Normalize(jobtype="estwrite", out_prefix="wgrptemplate_"),
                               name="warp2template",)

    get_bbox = setup_node(Function(function=get_bounding_box,
                                   input_names=["in_file"],
                                   output_names=["bbox"]),
                          name="get_bbox")

    # output
    pet_output = setup_node(IdentityInterface(fields=out_fields), name="pet_output")

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    wf.connect([
                # inputs
                (pet_input,   petpvc,  [("in_file", "pvc_input.in_file"),
                                        ("anat",    "pvc_input.reference_file"),
                                        ("tissues", "pvc_input.tissues")]),

                # get template bounding box to apply to results
                (pet_input, get_bbox,  [("pet_template", "in_file")]),

                # gunzip some inputs
                (pet_input, gunzip_pet,      [("in_file",      "in_file")]),
                (pet_input, gunzip_template, [("pet_template", "in_file")]),

                # gunzip some files for SPM Normalize
                (petpvc,    unzip_mrg, [("pvc_output.pvc_out",    "in1"),
                                        ("pvc_output.brain_mask", "in2"),
                                        ("pvc_output.gm_norm",    "in3")]),
                (pet_input, unzip_mrg, [("in_file",               "in4")]),

                (unzip_mrg, gunzipper, [("out", "in_file")]),

                (gunzipper, warp_mrg,  [("out_file", "in1")]),

                (warp_mrg, warp2template, [(("out", flatten_list), "apply_to_files")]),

                # prepare the target parameters of the warp to template
                (gunzip_pet,      warp2template, [("out_file", "source")]),
                (gunzip_template, warp2template, [("out_file", "template")]),
                (get_bbox,        warp2template, [("bbox",     "write_bounding_box")]),

                # output
                (warp2template, pet_output, [("normalization_parameters", "warp_field"),
                                             ("normalized_files" ,        "pvc_warped"),
                                             ("normalized_source",        "pet_warped"),
                                            ]),

                # output
                (petpvc,   pet_output, [("pvc_output.pvc_out",      "pvc_out"),
                                        ("pvc_output.brain_mask",   "brain_mask"),
                                        ("pvc_output.coreg_ref",    "coreg_ref"),
                                        ("pvc_output.coreg_others", "coreg_others"),
                                        ("pvc_output.gm_norm",      "gm_norm")]),
                ])

    if do_atlas:
        coreg_atlas = setup_node(spm_coregister(cost_function="mi"), name="coreg_atlas")

        # set the registration interpolation to nearest neighbour.
        coreg_atlas.inputs.write_interp = 0
        wf.connect([
                    (pet_input,   coreg_atlas, [("anat",                 "source")]),
                    (petpvc,      coreg_atlas, [("pvc_output.coreg_ref", "target")]),
                    (pet_input,   coreg_atlas, [("atlas_anat",           "apply_to_files")]),
                    (coreg_atlas, pet_output,  [("coregistered_files",   "atlas_pet")]),

                    # warp the atlas to the template space as well
                    (coreg_atlas, warp_mrg,    [("coregistered_files",   "in2")]),
        ])

    return wf


def attach_spm_mrpet_preprocessing(main_wf, wf_name="spm_mrpet_preproc",
                                   do_group_template=False):
    """ Attach a PET pre-processing workflow that uses SPM12 to `main_wf`.
    This workflow needs MRI based

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a
    `datasink` nodes.

    input_files.select.pet: input node

    datasink: nipype Node

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow

    do_group_template: bool
        If True will attach the group template creation and pre-processing pipeline.

    Nipype Workflow Dependencies
    ----------------------------
    This workflow depends on:
    - spm_anat_preproc

    Returns
    -------
    main_wf: nipype Workflow
    """
    # Dependency workflows
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)

    anat_output = get_interface_node(main_wf, "anat_output")

    # The base name of the 'pet' file for the substitutions
    anat_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'anat')))
    pet_fbasename  = remove_ext(op.basename(get_input_file_name(in_files, 'pet')))

    # get the PET preprocessing pipeline
    if do_group_template:
        pet_wf = spm_mrpet_grouptemplate_preprocessing(wf_name=wf_name)
        template_name = 'grptemplate'
        output_subfolder = 'grp_template'
    else:
        pet_wf = spm_mrpet_preprocessing(wf_name=wf_name)
        template_name = 'mni'
        output_subfolder = 'std_template'

    # dataSink output substitutions
    regexp_subst = [
                     (r"/{pet}_.*_pvc.nii.gz$",           "/{pet}_pvc.nii.gz"),
                     (r"/{pet}_.*_pvc_maths.nii.gz$",     "/{pet}_pvc_norm.nii.gz"),
                     (r"/{pet}_.*_pvc_intnormed.nii.gz$", "/{pet}_pvc_norm.nii.gz"),
                     (r"/tissues_brain_mask.nii$",        "/brain_mask_anat.nii"),
                     (r"/w{pet}.nii",                     "/{pet}_{template}.nii"),
                     (r"/w{pet}_.*_pvc.nii$",             "/{pet}_pvc_{template}.nii"),
                     (r"/w{pet}_.*_pvc_maths.nii$",       "/{pet}_pvc_norm_{template}.nii"),
                     (r"/w{pet}_.*_pvc_intnormed.nii$",   "/{pet}_pvc_norm_{template}.nii"),
                     (r"/wbrain_mask.nii",                "/brain_mask_{template}.nii"),
                     (r"/r{pet}.nii",                     "/{pet}_anat.nii"),
                     (r"/r{pet}_.*_pvc.nii$",             "/{pet}_pvc_anat.nii"),
                     (r"/r{pet}_.*_pvc_maths.nii$",       "/{pet}_pvc_norm_anat.nii"),
                     (r"/r{pet}_.*_pvc_intnormed.nii$",   "/{pet}_pvc_norm_anat.nii"),
                     (r"/y_rm{anat}_corrected.nii",       "/{anat}_{pet}_warpfield.nii"),
                     (r"/rm{anat}_corrected.nii$",        "/{anat}_{pet}.nii"),
                     (r"/rc1{anat}_corrected.nii$",       "/gm_{pet}.nii"),
                     (r"/rc2{anat}_corrected.nii$",       "/wm_{pet}.nii"),
                     (r"/rc3{anat}_corrected.nii$",       "/csf_{pet}.nii"),
                   ]
    regexp_subst = format_pair_list(regexp_subst, pet=pet_fbasename, anat=anat_fbasename,
                                    template=template_name)

    # prepare substitution for atlas_file, if any
    do_atlas, atlas_file = check_atlas_file()
    if do_atlas:
        atlas_basename = remove_ext(op.basename(atlas_file))
        regexp_subst.extend([
                             (r"/[\w]*{atlas}\.nii$", "/{atlas}_{pet}.nii"),
                            ])
        regexp_subst = format_pair_list(regexp_subst, pet=pet_fbasename, atlas=atlas_basename)

    regexp_subst += extension_duplicates(regexp_subst)

    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    # Connect the nodes
    main_wf.connect([
                     # pet file input
                     (in_files, pet_wf, [("pet", "pet_input.in_file")]),

                     # pet to anat registration
                     (anat_output,  pet_wf, [("anat_biascorr",  "pet_input.anat"),
                                             ("tissues_native", "pet_input.tissues"),
                                            ]),

                     (pet_wf, datasink, [
                                         ("pet_output.gm_norm",      "mrpet.@norm"),
                                         ("pet_output.coreg_others", "mrpet.tissues"),
                                         ("pet_output.coreg_ref",    "mrpet.@anat"),
                                         ("pet_output.pvc_mask",     "mrpet.@pvc_mask"),
                                         ("pet_output.pvc_out",      "mrpet.@pvc"),
                                         ("pet_output.brain_mask",   "mrpet.@brain_mask"),
                                         ("pet_output.pvc_warped",   "mrpet.{}.@pvc".format(output_subfolder)),
                                         ("pet_output.warp_field",   "mrpet.{}.@warp_field".format(output_subfolder)),
                                         ("pet_output.pet_warped",   "mrpet.{}.@pet_warped".format(output_subfolder)),
                                        ]),
                     ])

    if not do_group_template:
        # Connect the nodes
        main_wf.connect([
                         # pet to anat registration
                         (anat_output,  pet_wf, [("warp_forward", "pet_input.anat_to_mni_warp"),]),
                        ])


    if do_atlas:
            main_wf.connect([(anat_output, pet_wf,   [("atlas_anat",           "pet_input.atlas_anat")]),
                             (pet_wf,      datasink, [("pet_output.atlas_pet", "mrpet.@atlas")]),
                            ])

    return main_wf



# -*- coding: utf-8 -*-
"""
Nipype interfaces to canica and dictlearning in nilearn.decomposition
"""
import os.path as op

import numpy as np

from nipype.interfaces.base import (BaseInterface,
                                    BaseInterfaceInputSpec,
                                    TraitedSpec,
                                    InputMultiPath,
                                    OutputMultiPath,
                                    traits)
from nilearn.decomposition import CanICA, DictLearning

from ...utils import get_trait_value


class CanICAInputSpec(BaseInterfaceInputSpec):
    in_files = InputMultiPath(traits.File(desc="NifTI image file(s) from where to extract the data. \n"
                                               "If more than one, all should be spatially normalized.",
                                          exists=True, mandatory=True))
    mask = traits.File(desc="Mask to be used on data. If an instance of masker is passed, then its mask will be used.\n"
                            " If no mask is given, it will be computed automatically by a MultiNiftiMasker \n"
                            "with default parameters.",
                       exists=True)
    algorithm = traits.Enum(['canica', 'dictlearning'], desc="Desired nilearn ICA method.")
    confounds = traits.File(desc="CSV file path."
                                 "This parameter is passed to nilearn.signal.clean. "
                                 "Please see the related documentation for details.",
                            exists=True)
    n_components = traits.Int(desc="Number of components to extract.", default_value=20, usedefault=True)
    n_init = traits.Int(desc="CanICA only: The number of times the fastICA algorithm is restarted.", default_value=10, usedefault=True)
    n_epochs = traits.Int(desc="DictLearning only: Number of epochs the algorithm should run on the data.", default_value=1, usedefault=True)
    alpha = traits.Float(desc="DictLearning only: Sparsity controlling parameter.", default_value=1.0, usedefault=True)
    do_cca = traits.Bool(desc="CanICA only: Indicate if a Canonical Correlation Analysis must be run after the PCA.",
                         default_value=True, usedefault=True)
    memory = traits.Str(desc="Used to cache the masking process. By default, no caching is done. "
                        "If a string is given, it is the path to the caching directory.",
                        default_value='.', usedefault=True,)
    memory_level = traits.Int(desc="Rough estimator of the amount of memory used by caching. "
                              "Higher value means more memory for caching.", default_value=1, usedefault=True)
    threshold = traits.Either(traits.Bool, traits.Float,
                              desc="CanICA only: If None, no thresholding is applied.\n"
                                   "If ‘auto’, then we apply a thresholding that will keep the n_voxels, \n"
                                   "more intense voxels across all the maps, n_voxels being the number of voxels \n"
                                   "in a brain volume. A float value indicates the ratio of voxels to keep "
                                   "(2. means that the maps will together have 2 x n_voxels non-zero voxels ).",)
    random_state = traits.Int(desc="Pseudo number generator state used for random sampling.",)
    smoothing_fwhm = traits.Float(desc="If smoothing_fwhm is defined, it gives the full-width half maximum in "
                                       "millimeters of the spatial smoothing to apply to the signal.",)
    standardize = traits.Bool(desc="If standardize is True, the time-series are centered and normed: "
                                   "their mean is put to 0 and their variance to 1 in the time dimension.",
                              default_value=True, usedefault=True)
    n_jobs = traits.Int(desc="The number of CPUs to use to do the computation. -1 means 'all CPUs', "
                             "-2 'all CPUs but one', and so on.",
                        default_value=1, usedefault=True)


class CanICAOutputSpec(TraitedSpec):
    components = traits.File(desc="A nifti file with the reconstructed volume for each loading.")
    score      = traits.File(desc="Numpy txt file that holds the score for each subjects."
                                  "Score is two dimensional if per_component is True."
                                  "First dimension is squeezed if the number of subjects is one")
    loadings   = OutputMultiPath(traits.File(desc="For each subject, each sample, loadings for each "
                                                  "decomposition components shape: "
                                                  "number of subjects * (number of scans, number of regions))"))


class CanICAInterface(BaseInterface):
    """ Nipype Interface to NiLearn methods to perform Canonical Independent Component Analysis.

    For more information look at: nilearn.decomposition.CanICA
    """
    input_spec = CanICAInputSpec
    output_spec = CanICAOutputSpec

    def _run_interface(self, runtime):
        algorithm         = get_trait_value(self.inputs, 'algorithm',      default='canica')
        mask              = get_trait_value(self.inputs, 'mask')
        n_components      = get_trait_value(self.inputs, 'n_components')
        do_cca            = get_trait_value(self.inputs, 'do_cca')
        smoothing_fwhm    = get_trait_value(self.inputs, 'smoothing_fwhm', default=None)
        standardize       = get_trait_value(self.inputs, 'standardize',    default=None)
        threshold         = get_trait_value(self.inputs, 'threshold',      default=None)
        random_state      = get_trait_value(self.inputs, 'random_state',   default=None)
        n_init            = get_trait_value(self.inputs, 'n_init')
        n_jobs            = get_trait_value(self.inputs, 'n_jobs')
        n_epochs          = get_trait_value(self.inputs, 'n_epochs')
        alpha             = get_trait_value(self.inputs, 'alpha')
        memory            = get_trait_value(self.inputs, 'memory')
        memory_level      = get_trait_value(self.inputs, 'memory_level')
        confounds         = get_trait_value(self.inputs, 'confounds')

        # init the estimator
        if algorithm == 'canica':
            self._estimator = CanICA(mask=mask,
                                     n_components=n_components,
                                     threshold=threshold,
                                     random_state=random_state,
                                     standardize=standardize,
                                     smoothing_fwhm=smoothing_fwhm,
                                     do_cca=do_cca,
                                     verbose=1,
                                     n_init=n_init,
                                     memory=memory,
                                     memory_level=memory_level,
                                     n_jobs=n_jobs,
                                     )

        elif algorithm == 'dictlearning':
            self._estimator = DictLearning(mask=mask,
                                           n_components=n_components,
                                           random_state=random_state,
                                           standardize=standardize,
                                           smoothing_fwhm=smoothing_fwhm,
                                           verbose=1,
                                           n_epochs=n_epochs,
                                           alpha=alpha,
                                           memory=memory,
                                           memory_level=memory_level,
                                           n_jobs=n_jobs,
                                           )

        # set output file names
        self._estimator_name = algorithm
        self._confounds = confounds

        self._reconstructed_img_file = '{}_resting_state.nii.gz'.format(self._estimator_name)
        self._score_file   = '{}_score.txt'.format(self._estimator_name)
        self._loading_file = '{}_{}_loading.txt'

        # fit and transform
        self._estimator.fit(self.inputs.in_files, confounds=self._confounds)
        self._score    = self._estimator.score    (self.inputs.in_files, confounds=self._confounds)
        self._loadings = self._estimator.transform(self.inputs.in_files, confounds=self._confounds)

        return runtime

    def _list_outputs(self):
        outputs = self.output_spec().get()

        masker = self._estimator.masker_
        # Drop output maps to a Nifti file
        components_img = masker.inverse_transform(self._estimator.components_)
        components_img.to_filename(self._reconstructed_img_file)

        # save the score array
        if isinstance(self._score, float):
            with open(self._score_file, 'w') as f:
                f.write(str("%.10f" % self._score))
        else:
            np.savetxt(self._score_file, self._score, fmt='%.10f')

        # save the loadings files
        self._loading_files = []
        for idx, loadings in enumerate(self._loadings):
            loading_file = self._loading_file.format(self._estimator_name, idx)
            np.savetxt(loading_file, loadings, fmt='%.10f')
            self._loading_files.append(loading_file)

        outputs['components'] = op.abspath(self._reconstructed_img_file)
        outputs['score']      = op.abspath(self._score_file)
        outputs['loadings']   = [op.abspath(lf) for lf in self._loading_files]
        return outputs



from .decompose import attach_canica, attach_concat_canica

from .rsn_atlas import RestingStateNetworks

from .spatial_maps import (spatial_maps_goodness_of_fit,
                           spatial_maps_pairwise_similarity)

from .plotting import ( plot_connectivity_matrix,
                        plot_ica_results,
                        ica_loadings_sheet,
                        ICAResultsPlotter,
                        CanICAResultsPlotter,
                        MIALABICAResultsPlotter,
                        GIFTICAResultsPlotter,
                        GIFTGroupICAResultsPlotter,
                        SBMICAResultsPlotter,)

from .utils import (get_largest_blobs,
                    filter_ics,
                    add_groups_to_loadings_table,
                    build_raw_loadings_table)



# -*- coding: utf-8 -*-
"""
Nipype workflows to process resting-state functional MRI.
"""

from .grouptemplate import attach_spm_fmri_grouptemplate_wf
from .clean import attach_fmri_cleanup_wf
from .warp import attach_spm_warp_fmri_wf


def _attach_rest_preprocessing(main_wf, registration_wf_name="spm_warp_fmri", do_group_template=False):
    """ Attach the resting-state MRI pre-processing workflow to the `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    registration_wf_name: str
         Name of the registration workflow.

    do_group_template: bool
        If True will attach the group template creation and pre-processing pipeline.

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.select.anat: input node

    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    main_wf = attach_fmri_cleanup_wf(main_wf)
    main_wf = attach_spm_warp_fmri_wf(main_wf,
                                      registration_wf_name=registration_wf_name,
                                      do_group_template=False)

    if do_group_template:
        main_wf = attach_spm_fmri_grouptemplate_wf(main_wf, wf_name="spm_fmri_grptemplate")
        main_wf = attach_spm_warp_fmri_wf(main_wf,
                                          registration_wf_name=registration_wf_name,
                                          do_group_template=True)

        reg_wf       = main_wf.get_node("{}_{}".format(registration_wf_name, 'grptemplate'))
        grp_template = main_wf.get_node("group_template")
        main_wf.connect([(grp_template, reg_wf,  [("fmri_template",  "wfmri_input.epi_template")]),])

    return main_wf


def attach_rest_preprocessing(main_wf, wf_name="spm_warp_fmri"):
    return _attach_rest_preprocessing(main_wf,
                                      registration_wf_name=wf_name,
                                      do_group_template=False)


def attach_rest_grptemplate_preprocessing(main_wf, wf_name="spm_warp_fmri"):
    return _attach_rest_preprocessing(main_wf,
                                      registration_wf_name=wf_name,
                                      do_group_template=True)



# -*- coding: utf-8 -*-
"""
Utilities to use as mask and extract labels and ROIs from images.
"""
import nilearn.image as niimg
import numpy as np


def spread_labels(labels_img, roi_values=None, background_label=0):
    """ Spread each ROI in labels_img into a 4D image.

    Parameters
    ----------
    labels_img: 3D Niimg-like object
        See http://nilearn.github.io/manipulating_images/input_output.html. Region definitions, as one image of labels.

    roi_values: List of int or the values in the labels_img, optional
        The values of `labels_img` that you want to use.
        If None, will extract all the values except from `background_label`.

    background_label: number, optional
        Label used in labels_img to represent background.

    Return
    ------
    4d_labels_img: 4D Niimg-like object
    """
    lbls_img = niimg.load_img(labels_img)
    lbls_vol = lbls_img.get_data()

    if roi_values is None:
        roi_values = np.unique(lbls_vol)
        roi_values = roi_values[roi_values != background_label]

    atlas_roi_vols = [(lbls_vol == val) * val for val in roi_values]

    atlas_roi_imgs = [niimg.new_img_like(lbls_img, vol) for vol in atlas_roi_vols]
    # atlas_roi_imgs = [niimg.new_img_like(lbls_img, (lbls_vol == val) * val) for val in roi_values]

    return niimg.concat_imgs(atlas_roi_imgs)